import keras


class FeaturePyramid(keras.layers.Layer):
    """A Feature Pyramid Network (FPN) layer.

    This implements the paper:
        Tsung-Yi Lin, Piotr Dollar, Ross Girshick, Kaiming He, Bharath Hariharan,
        and Serge Belongie. Feature Pyramid Networks for Object Detection.
        (https://arxiv.org/pdf/1612.03144)

    Feature Pyramid Networks (FPNs) are basic components that are added to an
    existing feature extractor (CNN) to combine features at different scales.
    For the basic FPN, the inputs are features `Ci` from different levels of a
    CNN, which is usually the last block for each level, where the feature is
    scaled from the image by a factor of `1/2^i`.

    There is an output associated with each level in the basic FPN. The output
    Pi at level `i` (corresponding to Ci) is given by performing a merge
    operation on the outputs of:

    1) a lateral operation on Ci (usually a conv2D layer with kernel = 1 and
       strides = 1)
    2) a top-down upsampling operation from Pi+1 (except for the top most level)

    The final output of each level will also have a conv2D operation
    (typically with kernel = 3 and strides = 1).

    The inputs to the layer should be a dict with int keys should match the
    pyramid_levels, e.g. for `pyramid_levels` = [3,4,5], the expected input
    dict should be `{P3:c3, P4:c4, P5:c5}`.

    The output of the layer will have same structures as the inputs, a dict with
    extra coarser layers will be added based on the `max_level` provided.
    keys and value for each of the level.

    Args:
        min_level: int. The minimum level of the feature pyramid.
        max_level: int. The maximum level of the feature pyramid.
        num_filters: int. The number of filters in each feature map.
        activation: string or `keras.activations`. The activation function
            to be used in network.
            Defaults to `"relu"`.
        kernel_initializer: `str` or `keras.initializers` initializer.
            The kernel initializer for the convolution layers.
            Defaults to `"VarianceScaling"`.
        bias_initializer: `str` or `keras.initializers` initializer.
            The bias initializer for the convolution layers.
            Defaults to `"zeros"`.
        batch_norm_momentum: float.
            The momentum for the batch normalization layers.
            Defaults to `0.99`.
        batch_norm_epsilon: float.
            The epsilon for the batch normalization layers.
            Defaults to `0.001`.
        kernel_regularizer: `str` or `keras.regularizers` regularizer.
            The kernel regularizer for the convolution layers.
            Defaults to `None`.
        bias_regularizer: `str` or `keras.regularizers` regularizer.
            The bias regularizer for the convolution layers.
            Defaults to `None`.
        use_batch_norm: bool. Whether to use batch normalization.
            Defaults to `False`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `trainable`, `dtype` etc.
    """

    def __init__(
        self,
        min_level,
        max_level,
        num_filters=256,
        activation="relu",
        kernel_initializer="VarianceScaling",
        bias_initializer="zeros",
        batch_norm_momentum=0.99,
        batch_norm_epsilon=0.001,
        kernel_regularizer=None,
        bias_regularizer=None,
        use_batch_norm=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if min_level > max_level:
            raise ValueError(
                f"Minimum level ({min_level}) must be less than or equal to "
                f"maximum level ({max_level})."
            )
        self.min_level = min_level
        self.max_level = max_level
        self.num_filters = num_filters
        self.activation = keras.activations.get(activation)
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.batch_norm_momentum = batch_norm_momentum
        self.batch_norm_epsilon = batch_norm_epsilon
        self.use_batch_norm = use_batch_norm
        if kernel_regularizer is not None:
            self.kernel_regularizer = keras.regularizers.get(kernel_regularizer)
        else:
            self.kernel_regularizer = None
        if bias_regularizer is not None:
            self.bias_regularizer = keras.regularizers.get(bias_regularizer)
        else:
            self.bias_regularizer = None
        self.data_format = keras.backend.image_data_format()
        self.batch_norm_axis = -1 if self.data_format == "channels_last" else 1

    def build(self, input_shapes):
        input_shapes = {
            (
                input_name.split("_")[0]
                if "shape" in input_name
                else input_name
            ): input_shapes[input_name]
            for input_name in input_shapes
        }
        input_levels = [int(level[1]) for level in input_shapes]
        backbone_max_level = min(max(input_levels), self.max_level)

        # Build lateral layers
        self.lateral_conv_layers = {}
        for i in range(self.min_level, backbone_max_level + 1):
            level = f"P{i}"
            self.lateral_conv_layers[level] = keras.layers.Conv2D(
                filters=self.num_filters,
                kernel_size=1,
                padding="same",
                data_format=self.data_format,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                dtype=self.dtype_policy,
                name=f"lateral_conv_{level}",
            )
            self.lateral_conv_layers[level].build(input_shapes[level])

        self.lateral_batch_norm_layers = {}
        if self.use_batch_norm:
            for i in range(self.min_level, backbone_max_level + 1):
                level = f"P{i}"
                self.lateral_batch_norm_layers[level] = (
                    keras.layers.BatchNormalization(
                        axis=self.batch_norm_axis,
                        momentum=self.batch_norm_epsilon,
                        epsilon=self.batch_norm_epsilon,
                        name=f"lateral_norm_{level}",
                    )
                )
                self.lateral_batch_norm_layers[level].build(
                    (None, None, None, 256)
                    if self.data_format == "channels_last"
                    else (None, 256, None, None)
                )

        # Build output layers
        self.output_conv_layers = {}
        for i in range(self.min_level, backbone_max_level + 1):
            level = f"P{i}"
            self.output_conv_layers[level] = keras.layers.Conv2D(
                filters=self.num_filters,
                kernel_size=3,
                padding="same",
                data_format=self.data_format,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                dtype=self.dtype_policy,
                name=f"output_conv_{level}",
            )
            self.output_conv_layers[level].build(
                (None, None, None, 256)
                if self.data_format == "channels_last"
                else (None, 256, None, None)
            )

        # Build coarser layers
        for i in range(backbone_max_level + 1, self.max_level + 1):
            level = f"P{i}"
            self.output_conv_layers[level] = keras.layers.Conv2D(
                filters=self.num_filters,
                strides=2,
                kernel_size=3,
                padding="same",
                data_format=self.data_format,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                dtype=self.dtype_policy,
                name=f"coarser_{level}",
            )
            self.output_conv_layers[level].build(
                (None, None, None, 256)
                if self.data_format == "channels_last"
                else (None, 256, None, None)
            )

        # Build batch norm layers
        self.output_batch_norms = {}
        if self.use_batch_norm:
            for i in range(self.min_level, self.max_level + 1):
                level = f"P{i}"
                self.output_batch_norms[level] = (
                    keras.layers.BatchNormalization(
                        axis=self.batch_norm_axis,
                        momentum=self.batch_norm_epsilon,
                        epsilon=self.batch_norm_epsilon,
                        name=f"output_norm_{level}",
                    )
                )
                self.output_batch_norms[level].build(
                    (None, None, None, 256)
                    if self.data_format == "channels_last"
                    else (None, 256, None, None)
                )

        # The same upsampling layer is used for all levels
        self.top_down_op = keras.layers.UpSampling2D(
            size=2,
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="upsampling",
        )
        # The same merge layer is used for all levels
        self.merge_op = keras.layers.Add(
            dtype=self.dtype_policy, name="merge_op"
        )

        self.built = True

    def call(self, inputs):
        """
        Inputs:
            The input to the model is expected to be an `Dict[Tensors]`,
                containing the feature maps on top of which the FPN
                will be added.

        Outputs:
            A dictionary of feature maps and added coarser levels based
                on minimum and maximum levels provided to the layer.
        """

        output_features = {}

        # Get the backbone max level
        input_levels = [int(level[1]) for level in inputs]
        backbone_max_level = min(max(input_levels), self.max_level)

        for i in range(backbone_max_level, self.min_level - 1, -1):
            level = f"P{i}"
            output = self.lateral_conv_layers[level](inputs[level])
            if i < backbone_max_level:
                # for the top most output, it doesn't need to merge with any
                # upper stream outputs
                upstream_output = self.top_down_op(output_features[f"P{i+1}"])
                output = self.merge_op([output, upstream_output])
            output_features[level] = (
                self.lateral_batch_norm_layers[level](output)
                if self.use_batch_norm
                else output
            )

        # Post apply the output layers so that we don't leak them to the down
        # stream level
        for i in range(backbone_max_level, self.min_level - 1, -1):
            level = f"P{i}"
            output_features[level] = self.output_conv_layers[level](
                output_features[level]
            )

        for i in range(backbone_max_level + 1, self.max_level + 1):
            level = f"P{i}"
            feats_in = output_features[f"P{i-1}"]
            if i > backbone_max_level + 1:
                feats_in = self.activation(feats_in)
            output_features[level] = (
                self.output_batch_norms[level](
                    self.output_conv_layers[level](feats_in)
                )
                if self.use_batch_norm
                else self.output_conv_layers[level](feats_in)
            )

        return output_features

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "min_level": self.min_level,
                "max_level": self.max_level,
                "num_filters": self.num_filters,
                "use_batch_norm": self.use_batch_norm,
                "activation": keras.activations.serialize(self.activation),
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "batch_norm_momentum": self.batch_norm_momentum,
                "batch_norm_epsilon": self.batch_norm_epsilon,
                "kernel_regularizer": (
                    keras.regularizers.serialize(self.kernel_regularizer)
                    if self.kernel_regularizer is not None
                    else None
                ),
                "bias_regularizer": (
                    keras.regularizers.serialize(self.bias_regularizer)
                    if self.bias_regularizer is not None
                    else None
                ),
            }
        )

        return config

    def compute_output_shape(self, input_shapes):
        output_shape = {}
        print(input_shapes)
        input_levels = [int(level[1]) for level in input_shapes]
        backbone_max_level = min(max(input_levels), self.max_level)

        for i in range(self.min_level, backbone_max_level + 1):
            level = f"P{i}"
            if self.data_format == "channels_last":
                output_shape[level] = input_shapes[level][:-1] + (256,)
            else:
                output_shape[level] = (
                    input_shapes[level][0],
                    256,
                ) + input_shapes[level][1:3]

        intermediate_shape = input_shapes[f"P{backbone_max_level}"]
        intermediate_shape = (
            (
                intermediate_shape[0],
                intermediate_shape[1] // 2,
                intermediate_shape[2] // 2,
                256,
            )
            if self.data_format == "channels_last"
            else (
                intermediate_shape[0],
                256,
                intermediate_shape[1] // 2,
                intermediate_shape[2] // 2,
            )
        )

        for i in range(backbone_max_level + 1, self.max_level + 1):
            level = f"P{i}"
            output_shape[level] = intermediate_shape
            intermediate_shape = (
                (
                    intermediate_shape[0],
                    intermediate_shape[1] // 2,
                    intermediate_shape[2] // 2,
                    256,
                )
                if self.data_format == "channels_last"
                else (
                    intermediate_shape[0],
                    256,
                    intermediate_shape[1] // 2,
                    intermediate_shape[2] // 2,
                )
            )

        return output_shape



from keras_hub.src.models.f_net.f_net_backbone import FNetBackbone
from keras_hub.src.models.f_net.f_net_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, FNetBackbone)



import math

import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.efficientnet.fusedmbconv import FusedMBConvBlock
from keras_hub.src.models.efficientnet.mbconv import MBConvBlock
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone


@keras_hub_export("keras_hub.models.EfficientNetBackbone")
class EfficientNetBackbone(FeaturePyramidBackbone):
    """An EfficientNet backbone model.

    This class encapsulates the architectures for both EfficientNetV1 and
    EfficientNetV2. EfficientNetV2 uses Fused-MBConv Blocks and Neural
    Architecture Search (NAS) to make models sizes much smaller while still
    improving overall model quality.

    References:
    - [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks]
        (https://arxiv.org/abs/1905.11946) (ICML 2019)
    - [Based on the original keras.applications EfficientNet]
        (https://github.com/keras-team/keras/blob/master/keras/applications/efficientnet.py)
    - [EfficientNetV2: Smaller Models and Faster Training]
        (https://arxiv.org/abs/2104.00298) (ICML 2021)

    Args:
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        dropout: float, dropout rate at skip connections. The default
            value is set to 0.2.
        depth_divisor: integer, a unit of network width. The default value is
            set to 8.
        activation: activation function to use between each convolutional layer.
        input_shape: optional shape tuple, it should have exactly 3 input
            channels.
        stackwise_kernel_sizes:  list of ints, the kernel sizes used for each
            conv block.
        stackwise_num_repeats: list of ints, number of times to repeat each
            conv block.
        stackwise_input_filters: list of ints, number of input filters for
            each conv block.
        stackwise_output_filters: list of ints, number of output filters for
            each stack in the conv blocks model.
        stackwise_expansion_ratios: list of floats, expand ratio passed to the
            squeeze and excitation blocks.
        stackwise_strides: list of ints, stackwise_strides for each conv block.
        stackwise_squeeze_and_excite_ratios: list of ints, the squeeze and
            excite ratios passed to the squeeze and excitation blocks.
        stackwise_block_types: list of strings.  Each value is either 'v1',
            'unfused' or 'fused' depending on the desired blocks.  'v1' uses the
            original efficientnet block. FusedMBConvBlock is similar to
            MBConvBlock, but instead of using a depthwise convolution and a 1x1
            output convolution blocks fused blocks use a single 3x3 convolution
            block.
        min_depth: integer, minimum number of filters. Can be None and ignored
            if use_depth_divisor_as_min_depth is set to True.
        include_initial_padding: bool, whether to include initial zero padding
            (as per v1).
        use_depth_divisor_as_min_depth: bool, whether to use depth_divisor as
            the minimum depth instead of min_depth (as per v1).
        cap_round_filter_decrease: bool, whether to cap the max decrease in the
            number of filters the rounding process potentially produces
            (as per v1).
        stem_conv_padding: str, can be 'same' or 'valid'. Padding for the stem.
        batch_norm_momentum: float, momentum for the moving average calcualtion
            in the batch normalization layers.

    Example:
    ```python
    # You can customize the EfficientNet architecture:
    model = EfficientNetBackbone(
        stackwise_kernel_sizes=[3, 3, 3, 3, 3, 3],
        stackwise_num_repeats=[2, 4, 4, 6, 9, 15],
        stackwise_input_filters=[24, 24, 48, 64, 128, 160],
        stackwise_output_filters=[24, 48, 64, 128, 160, 256],
        stackwise_expansion_ratios=[1, 4, 4, 4, 6, 6],
        stackwise_squeeze_and_excite_ratios=[0.0, 0.0, 0, 0.25, 0.25, 0.25],
        stackwise_strides=[1, 2, 2, 2, 1, 2],
        stackwise_block_types=[["fused"] * 3 + ["unfused"] * 3],
        width_coefficient=1.0,
        depth_coefficient=1.0,
    )
    images = np.ones((1, 256, 256, 3))
    outputs = efficientnet.predict(images)
    ```
    """

    def __init__(
        self,
        *,
        width_coefficient,
        depth_coefficient,
        stackwise_kernel_sizes,
        stackwise_num_repeats,
        stackwise_input_filters,
        stackwise_output_filters,
        stackwise_expansion_ratios,
        stackwise_squeeze_and_excite_ratios,
        stackwise_strides,
        stackwise_block_types,
        dropout=0.2,
        depth_divisor=8,
        min_depth=8,
        input_shape=(None, None, 3),
        activation="swish",
        include_initial_padding=False,
        use_depth_divisor_as_min_depth=False,
        cap_round_filter_decrease=False,
        stem_conv_padding="same",
        batch_norm_momentum=0.9,
        **kwargs,
    ):
        image_input = keras.layers.Input(shape=input_shape)

        x = image_input  # Intermediate result.
        if include_initial_padding:
            x = keras.layers.ZeroPadding2D(
                padding=self._correct_pad_downsample(x, 3),
                name="stem_conv_pad",
            )(x)

        # Build stem
        stem_filters = round_filters(
            filters=stackwise_input_filters[0],
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor,
            use_depth_divisor_as_min_depth=use_depth_divisor_as_min_depth,
            cap_round_filter_decrease=cap_round_filter_decrease,
        )

        x = keras.layers.Conv2D(
            filters=stem_filters,
            kernel_size=3,
            strides=2,
            padding=stem_conv_padding,
            use_bias=False,
            kernel_initializer=conv_kernel_initializer(),
            name="stem_conv",
        )(x)

        x = keras.layers.BatchNormalization(
            momentum=batch_norm_momentum,
            name="stem_bn",
        )(x)
        x = keras.layers.Activation(activation, name="stem_activation")(x)

        # Build blocks
        block_id = 0
        blocks = float(sum(stackwise_num_repeats))

        self._pyramid_outputs = {}
        curr_pyramid_level = 1

        for i in range(len(stackwise_kernel_sizes)):
            num_repeats = stackwise_num_repeats[i]
            input_filters = stackwise_input_filters[i]
            output_filters = stackwise_output_filters[i]

            # Update block input and output filters based on depth multiplier.
            input_filters = round_filters(
                filters=input_filters,
                width_coefficient=width_coefficient,
                min_depth=min_depth,
                depth_divisor=depth_divisor,
                use_depth_divisor_as_min_depth=use_depth_divisor_as_min_depth,
                cap_round_filter_decrease=cap_round_filter_decrease,
            )
            output_filters = round_filters(
                filters=output_filters,
                width_coefficient=width_coefficient,
                min_depth=min_depth,
                depth_divisor=depth_divisor,
                use_depth_divisor_as_min_depth=use_depth_divisor_as_min_depth,
                cap_round_filter_decrease=cap_round_filter_decrease,
            )

            repeats = round_repeats(
                repeats=num_repeats,
                depth_coefficient=depth_coefficient,
            )
            strides = stackwise_strides[i]
            squeeze_and_excite_ratio = stackwise_squeeze_and_excite_ratios[i]

            for j in range(repeats):
                # The first block needs to take care of stride and filter size
                # increase.
                if j > 0:
                    strides = 1
                    input_filters = output_filters

                if strides != 1:
                    self._pyramid_outputs[f"P{curr_pyramid_level}"] = x
                    curr_pyramid_level += 1

                # 97 is the start of the lowercase alphabet.
                letter_identifier = chr(j + 97)
                stackwise_block_type = stackwise_block_types[i]
                block_name = f"block{i + 1}{letter_identifier}_"
                if stackwise_block_type == "v1":
                    x = self._apply_efficientnet_block(
                        inputs=x,
                        filters_in=input_filters,
                        filters_out=output_filters,
                        kernel_size=stackwise_kernel_sizes[i],
                        strides=strides,
                        expand_ratio=stackwise_expansion_ratios[i],
                        se_ratio=squeeze_and_excite_ratio,
                        activation=activation,
                        dropout=dropout * block_id / blocks,
                        name=block_name,
                    )
                else:
                    block = get_conv_constructor(stackwise_block_type)(
                        input_filters=input_filters,
                        output_filters=output_filters,
                        expand_ratio=stackwise_expansion_ratios[i],
                        kernel_size=stackwise_kernel_sizes[i],
                        strides=strides,
                        se_ratio=squeeze_and_excite_ratio,
                        activation=activation,
                        dropout=dropout * block_id / blocks,
                        batch_norm_momentum=batch_norm_momentum,
                        name=block_name,
                    )
                    x = block(x)
                block_id += 1

        # Build top
        top_filters = round_filters(
            filters=1280,
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor,
            use_depth_divisor_as_min_depth=use_depth_divisor_as_min_depth,
            cap_round_filter_decrease=cap_round_filter_decrease,
        )

        x = keras.layers.Conv2D(
            filters=top_filters,
            kernel_size=1,
            padding="same",
            strides=1,
            kernel_initializer=conv_kernel_initializer(),
            use_bias=False,
            name="top_conv",
            data_format="channels_last",
        )(x)
        x = keras.layers.BatchNormalization(
            momentum=batch_norm_momentum,
            name="top_bn",
        )(x)
        x = keras.layers.Activation(
            activation=activation, name="top_activation"
        )(x)

        self._pyramid_outputs[f"P{curr_pyramid_level}"] = x
        curr_pyramid_level += 1

        # Create model.
        super().__init__(inputs=image_input, outputs=x, **kwargs)

        # === Config ===
        self.width_coefficient = width_coefficient
        self.depth_coefficient = depth_coefficient
        self.dropout = dropout
        self.depth_divisor = depth_divisor
        self.min_depth = min_depth
        self.activation = activation
        self.stackwise_kernel_sizes = stackwise_kernel_sizes
        self.stackwise_num_repeats = stackwise_num_repeats
        self.stackwise_input_filters = stackwise_input_filters
        self.stackwise_output_filters = stackwise_output_filters
        self.stackwise_expansion_ratios = stackwise_expansion_ratios
        self.stackwise_squeeze_and_excite_ratios = (
            stackwise_squeeze_and_excite_ratios
        )
        self.stackwise_strides = stackwise_strides
        self.stackwise_block_types = stackwise_block_types

        self.include_initial_padding = include_initial_padding
        self.use_depth_divisor_as_min_depth = use_depth_divisor_as_min_depth
        self.cap_round_filter_decrease = cap_round_filter_decrease
        self.stem_conv_padding = stem_conv_padding
        self.batch_norm_momentum = batch_norm_momentum

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width_coefficient": self.width_coefficient,
                "depth_coefficient": self.depth_coefficient,
                "dropout": self.dropout,
                "depth_divisor": self.depth_divisor,
                "min_depth": self.min_depth,
                "activation": self.activation,
                "input_shape": self.input_shape[1:],
                "stackwise_kernel_sizes": self.stackwise_kernel_sizes,
                "stackwise_num_repeats": self.stackwise_num_repeats,
                "stackwise_input_filters": self.stackwise_input_filters,
                "stackwise_output_filters": self.stackwise_output_filters,
                "stackwise_expansion_ratios": self.stackwise_expansion_ratios,
                "stackwise_squeeze_and_excite_ratios": self.stackwise_squeeze_and_excite_ratios,
                "stackwise_strides": self.stackwise_strides,
                "stackwise_block_types": self.stackwise_block_types,
                "include_initial_padding": self.include_initial_padding,
                "use_depth_divisor_as_min_depth": self.use_depth_divisor_as_min_depth,
                "cap_round_filter_decrease": self.cap_round_filter_decrease,
                "stem_conv_padding": self.stem_conv_padding,
                "batch_norm_momentum": self.batch_norm_momentum,
            }
        )
        return config

    def _correct_pad_downsample(self, inputs, kernel_size):
        """Returns a tuple for zero-padding for 2D convolution with downsampling.

        Args:
            inputs: Input tensor.
            kernel_size: An integer or tuple/list of 2 integers.

        Returns:
            A tuple.
        """
        img_dim = 1
        input_size = inputs.shape[img_dim : (img_dim + 2)]
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if input_size[0] is None:
            adjust = (1, 1)
        else:
            adjust = (1 - input_size[0] % 2, 1 - input_size[1] % 2)
        correct = (kernel_size[0] // 2, kernel_size[1] // 2)
        return (
            (correct[0] - adjust[0], correct[0]),
            (correct[1] - adjust[1], correct[1]),
        )

    def _apply_efficientnet_block(
        self,
        inputs,
        filters_in=32,
        filters_out=16,
        kernel_size=3,
        strides=1,
        activation="swish",
        expand_ratio=1,
        se_ratio=0.0,
        dropout=0.0,
        name="",
    ):
        """An inverted residual block.

        Args:
            inputs: Tensor, The input tensor of the block
            filters_in: integer, the number of input filters.
            filters_out: integer, the number of output filters.
            kernel_size: integer, the dimension of the convolution window.
            strides: integer, the stride of the convolution.
            activation: activation function to use between each convolutional layer.
            expand_ratio: integer, scaling coefficient for the input filters.
            se_ratio: float between 0 and 1, fraction to squeeze the input filters.
            dropout: float between 0 and 1, fraction of the input units to drop.
            name: string, block label.

        Returns:
            output tensor for the block.
        """
        filters = filters_in * expand_ratio
        if expand_ratio != 1:
            x = keras.layers.Conv2D(
                filters=filters,
                kernel_size=1,
                strides=1,
                padding="same",
                use_bias=False,
                kernel_initializer=conv_kernel_initializer(),
                name=name + "expand_conv",
            )(inputs)
            x = keras.layers.BatchNormalization(
                axis=3,
                name=name + "expand_bn",
            )(x)
            x = keras.layers.Activation(
                activation, name=name + "expand_activation"
            )(x)
        else:
            x = inputs

        # Depthwise Convolution
        if strides == 2:
            x = keras.layers.ZeroPadding2D(
                padding=self._correct_pad_downsample(x, kernel_size),
                name=name + "dwconv_pad",
            )(x)
            conv_pad = "valid"
        else:
            conv_pad = "same"

        x = keras.layers.DepthwiseConv2D(
            kernel_size=kernel_size,
            strides=strides,
            padding=conv_pad,
            use_bias=False,
            depthwise_initializer=conv_kernel_initializer(),
            name=name + "dwconv",
        )(x)
        x = keras.layers.BatchNormalization(
            axis=3,
            name=name + "dwconv_bn",
        )(x)
        x = keras.layers.Activation(
            activation, name=name + "dwconv_activation"
        )(x)

        # Squeeze and Excitation phase
        if 0 < se_ratio <= 1:
            filters_se = max(1, int(filters_in * se_ratio))
            se = keras.layers.GlobalAveragePooling2D(name=name + "se_squeeze")(
                x
            )
            se_shape = (1, 1, filters)
            se = keras.layers.Reshape(se_shape, name=name + "se_reshape")(se)
            se = keras.layers.Conv2D(
                filters_se,
                1,
                padding="same",
                activation=activation,
                kernel_initializer=conv_kernel_initializer(),
                name=name + "se_reduce",
            )(se)
            se = keras.layers.Conv2D(
                filters,
                1,
                padding="same",
                activation="sigmoid",
                kernel_initializer=conv_kernel_initializer(),
                name=name + "se_expand",
            )(se)
            x = keras.layers.multiply([x, se], name=name + "se_excite")

        # Output phase
        x = keras.layers.Conv2D(
            filters=filters_out,
            kernel_size=1,
            strides=1,
            padding="same",
            use_bias=False,
            kernel_initializer=conv_kernel_initializer(),
            name=name + "project",
        )(x)
        x = keras.layers.BatchNormalization(
            axis=3,
            name=name + "project_bn",
        )(x)
        x = keras.layers.Activation(
            activation, name=name + "project_activation"
        )(x)

        if strides == 1 and filters_in == filters_out:
            if dropout > 0:
                x = keras.layers.Dropout(
                    dropout,
                    noise_shape=(None, 1, 1, 1),
                    name=name + "drop",
                )(x)
            x = keras.layers.Add(name=name + "add")([x, inputs])

        return x


def conv_kernel_initializer(scale=2.0):
    return keras.initializers.VarianceScaling(
        scale=scale, mode="fan_out", distribution="truncated_normal"
    )


def round_filters(
    filters,
    width_coefficient,
    min_depth,
    depth_divisor,
    use_depth_divisor_as_min_depth,
    cap_round_filter_decrease,
):
    """Round number of filters based on depth multiplier.

    Args:
        filters: int, number of filters for Conv layer
        width_coefficient: float, denotes the scaling coefficient of network
            width
        depth_divisor: int, a unit of network width
        use_depth_divisor_as_min_depth: bool, whether to use depth_divisor as
            the minimum depth instead of min_depth (as per v1)
        max_round_filter_decrease: bool, whether to cap the decrease in the
            number of filters this process produces (as per v1)

    Returns:
        int, new rounded filters value for Conv layer
    """
    filters *= width_coefficient

    if use_depth_divisor_as_min_depth:
        min_depth = depth_divisor

    new_filters = max(
        min_depth,
        int(filters + depth_divisor / 2) // depth_divisor * depth_divisor,
    )

    if cap_round_filter_decrease:
        # Make sure that round down does not go down by more than 10%.
        if new_filters < 0.9 * filters:
            new_filters += depth_divisor

    return int(new_filters)


def round_repeats(repeats, depth_coefficient):
    """Round number of repeats based on depth multiplier.

    Args:
        repeats: int, number of repeats of efficientnet block
        depth_coefficient: float, denotes the scaling coefficient of network
            depth

    Returns:
        int, rounded repeats
    """
    return int(math.ceil(depth_coefficient * repeats))


def get_conv_constructor(conv_type):
    if conv_type == "unfused":
        return MBConvBlock
    elif conv_type == "fused":
        return FusedMBConvBlock
    else:
        raise ValueError(
            "Expected `conv_type` to be "
            "one of 'unfused', 'fused', but got "
            f"`conv_type={conv_type}`"
        )



"""Llama model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "llama2_7b_en": {
        "metadata": {
            "description": "7 billion parameter, 32-layer, base LLaMA 2 model.",
            "params": 6738415616,
            "official_name": "LLaMA 2",
            "path": "llama2",
            "model_card": "https://github.com/meta-llama/llama",
        },
        "kaggle_handle": "kaggle://keras/llama2/keras/llama2_7b_en/1",
    },
    "llama2_7b_en_int8": {
        "metadata": {
            "description": (
                "7 billion parameter, 32-layer, base LLaMA 2 model with "
                "activation and weights quantized to int8."
            ),
            "params": 6739839488,
            "official_name": "LLaMA 2",
            "path": "llama2",
            "model_card": "https://github.com/meta-llama/llama",
        },
        "kaggle_handle": "kaggle://keras/llama2/keras/llama2_7b_en_int8/1",
    },
    "llama2_instruct_7b_en": {
        "metadata": {
            "description": (
                "7 billion parameter, 32-layer, instruction tuned LLaMA 2 "
                "model."
            ),
            "params": 6738415616,
            "official_name": "LLaMA 2",
            "path": "llama2",
            "model_card": "https://github.com/meta-llama/llama",
        },
        "kaggle_handle": "kaggle://keras/llama2/keras/llama2_instruct_7b_en/1",
    },
    "llama2_instruct_7b_en_int8": {
        "metadata": {
            "description": (
                "7 billion parameter, 32-layer, instruction tuned LLaMA 2 "
                "model with activation and weights quantized to int8."
            ),
            "params": 6739839488,
            "official_name": "LLaMA 2",
            "path": "llama2",
            "model_card": "https://github.com/meta-llama/llama",
        },
        "kaggle_handle": "kaggle://keras/llama2/keras/llama2_instruct_7b_en_int8/1",
    },
    "vicuna_1.5_7b_en": {
        "metadata": {
            "description": (
                "7 billion parameter, 32-layer, instruction tuned Vicuna v1.5 "
                "model."
            ),
            "params": 6738415616,
            "official_name": "Vicuna",
            "path": "vicuna",
            "model_card": "https://github.com/lm-sys/FastChat",
        },
        "kaggle_handle": "kaggle://keras/vicuna/keras/vicuna_1.5_7b_en/1",
    },
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MiT model preset configurations."""

backbone_presets_with_weights = {
    "mit_b0_ade20k_512": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 3321962,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b0_ade20k_512/1",
    },
    "mit_b1_ade20k_512": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 13156554,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b1_ade20k_512/1",
    },
    "mit_b2_ade20k_512": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 16 transformer blocks."
            ),
            "params": 24201418,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b2_ade20k_512/1",
    },
    "mit_b3_ade20k_512": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 28 transformer blocks."
            ),
            "params": 44077258,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b3_ade20k_512/1",
    },
    "mit_b4_ade20k_512": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 41 transformer blocks."
            ),
            "params": 60847818,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b4_ade20k_512/1",
    },
    "mit_b5_ade20k_640": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 52 transformer blocks."
            ),
            "params": 81448138,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b5_ade20k_512/1",
    },
    "mit_b0_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 3321962,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b0_cityscapes_1024/1",
    },
    "mit_b1_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 13156554,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b1_cityscapes_1024/1",
    },
    "mit_b2_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 16 transformer blocks."
            ),
            "params": 24201418,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b2_cityscapes_1024/1",
    },
    "mit_b3_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 28 transformer blocks."
            ),
            "params": 44077258,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b3_cityscapes_1024/1",
    },
    "mit_b4_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 41 transformer blocks."
            ),
            "params": 60847818,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b4_cityscapes_1024/1",
    },
    "mit_b5_cityscapes_1024": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 52 transformer blocks."
            ),
            "params": 81448138,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b5_cityscapes_1024/1",
    },
}

backbone_presets = {
    **backbone_presets_with_weights,
}



from keras_hub.src.models.bart.bart_backbone import BartBackbone
from keras_hub.src.models.bart.bart_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, BartBackbone)



from keras_hub.src.models.pali_gemma.pali_gemma_backbone import (
    PaliGemmaBackbone,
)
from keras_hub.src.models.pali_gemma.pali_gemma_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, PaliGemmaBackbone)



import keras
import numpy as np
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone
from keras_hub.src.models.mit.mit_layers import HierarchicalTransformerEncoder
from keras_hub.src.models.mit.mit_layers import OverlappingPatchingAndEmbedding


@keras_hub_export("keras_hub.models.MiTBackbone")
class MiTBackbone(FeaturePyramidBackbone):
    def __init__(
        self,
        depths,
        num_layers,
        blockwise_num_heads,
        blockwise_sr_ratios,
        max_drop_path_rate,
        patch_sizes,
        strides,
        image_shape=(None, None, 3),
        hidden_dims=None,
        **kwargs,
    ):
        """A Backbone implementing the MixTransformer.

        This architecture to be used as a backbone for the SegFormer
        architecture [SegFormer: Simple and Efficient Design for Semantic
        Segmentation with Transformers](https://arxiv.org/abs/2105.15203)
        [Based on the TensorFlow implementation from DeepVision](
            https://github.com/DavidLandup0/deepvision/tree/main/deepvision/models/classification/mix_transformer)

        Args:
            depths: The number of transformer encoders to be used per layer in the
                network.
            num_layers: int. The number of Transformer layers.
            blockwise_num_heads: list of integers, the number of heads to use
            in the attention computation for each layer.
            blockwise_sr_ratios: list of integers, the sequence reduction
                ratio to perform for each layer on the sequence before key and
                value projections. If set to > 1, a `Conv2D` layer is used to
                reduce the length of the sequence.
            max_drop_path_rate: The final value of the `linspace()` that
                defines the drop path rates for the `DropPath` layers of
                the `HierarchicalTransformerEncoder` layers.
            image_shape: optional shape tuple, defaults to (None, None, 3).
            hidden_dims: the embedding dims per hierarchical layer, used as
                the levels of the feature pyramid.
            patch_sizes: list of integers, the patch_size to apply for each layer.
            strides: list of integers, stride to apply for each layer.

        Examples:

        Using the class with a `backbone`:

        ```python
        images = np.ones(shape=(1, 96, 96, 3))
        labels = np.zeros(shape=(1, 96, 96, 1))
        backbone = keras_hub.models.MiTBackbone.from_preset("mit_b0_ade20k_512")

        # Evaluate model
        model(images)

        # Train model
        model.compile(
            optimizer="adam",
            loss=keras.losses.BinaryCrossentropy(from_logits=False),
            metrics=["accuracy"],
        )
        model.fit(images, labels, epochs=3)
        ```
        """
        dpr = [x for x in np.linspace(0.0, max_drop_path_rate, sum(depths))]

        # === Layers ===
        cur = 0
        patch_embedding_layers = []
        transformer_blocks = []
        layer_norms = []

        for i in range(num_layers):
            patch_embed_layer = OverlappingPatchingAndEmbedding(
                project_dim=hidden_dims[i],
                patch_size=patch_sizes[i],
                stride=strides[i],
                name=f"patch_and_embed_{i}",
            )
            patch_embedding_layers.append(patch_embed_layer)

            transformer_block = [
                HierarchicalTransformerEncoder(
                    project_dim=hidden_dims[i],
                    num_heads=blockwise_num_heads[i],
                    sr_ratio=blockwise_sr_ratios[i],
                    drop_prob=dpr[cur + k],
                    name=f"hierarchical_encoder_{i}_{k}",
                )
                for k in range(depths[i])
            ]
            transformer_blocks.append(transformer_block)
            cur += depths[i]
            layer_norms.append(keras.layers.LayerNormalization())

        # === Functional Model ===
        image_input = keras.layers.Input(shape=image_shape)
        x = image_input  # Intermediate result.
        pyramid_outputs = {}
        for i in range(num_layers):
            # Compute new height/width after the `proj`
            # call in `OverlappingPatchingAndEmbedding`
            stride = strides[i]
            new_height, new_width = (
                int(ops.shape(x)[1] / stride),
                int(ops.shape(x)[2] / stride),
            )

            x = patch_embedding_layers[i](x)
            for blk in transformer_blocks[i]:
                x = blk(x)
            x = layer_norms[i](x)
            x = keras.layers.Reshape(
                (new_height, new_width, -1), name=f"output_level_{i}"
            )(x)
            pyramid_outputs[f"P{i + 1}"] = x

        super().__init__(inputs=image_input, outputs=x, **kwargs)

        # === Config ===
        self.depths = depths
        self.image_shape = image_shape
        self.hidden_dims = hidden_dims
        self.pyramid_outputs = pyramid_outputs
        self.num_layers = num_layers
        self.blockwise_num_heads = blockwise_num_heads
        self.blockwise_sr_ratios = blockwise_sr_ratios
        self.max_drop_path_rate = max_drop_path_rate
        self.patch_sizes = patch_sizes
        self.strides = strides

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "depths": self.depths,
                "hidden_dims": self.hidden_dims,
                "image_shape": self.image_shape,
                "num_layers": self.num_layers,
                "blockwise_num_heads": self.blockwise_num_heads,
                "blockwise_sr_ratios": self.blockwise_sr_ratios,
                "max_drop_path_rate": self.max_drop_path_rate,
                "patch_sizes": self.patch_sizes,
                "strides": self.strides,
            }
        )
        return config



import math

import keras
from keras import ops

from keras_hub.src.bounding_box.converters import convert_format


class AnchorGenerator(keras.layers.Layer):
    """Generates anchor boxes for object detection tasks.

    This layer creates a set of anchor boxes (also known as default boxes or
    priors) for use in object detection models, particularly those utilizing
    Feature Pyramid Networks (FPN). It generates anchors across multiple
    pyramid levels, with various scales and aspect ratios.

    Feature Pyramid Levels:
    - Levels typically range from 2 to 6 (P2 to P7), corresponding to different
        resolutions of the input image.
    - Each level l has a stride of 2^l pixels relative to the input image.
    - Lower levels (e.g., P2) have higher resolution and are used for
        detecting smaller objects.
    - Higher levels (e.g., P7) have lower resolution and are used
        for larger objects.

    Args:
        bounding_box_format: str. The format of the bounding boxes
            to be generated. Expected to be a string like 'xyxy', 'xywh', etc.
        min_level: int. Minimum level of the output feature pyramid.
        max_level: int. Maximum level of the output feature pyramid.
        num_scales: int. Number of intermediate scales added on each level.
            For example, num_scales=2 adds one additional intermediate anchor
            scale [2^0, 2^0.5] on each level.
        aspect_ratios:  List[float]. Aspect ratios of anchors added on
            each level. Each number indicates the ratio of width to height.
        anchor_size: float. Scale of size of the base anchor relative to the
            feature stride 2^level.

    Call arguments:
        inputs: An image tensor with shape `[B, H, W, C]` or
            `[H, W, C]`. Its shape will be used to determine anchor
            sizes.

    Returns:
        Dict: A dictionary mapping feature levels
            (e.g., 'P3', 'P4', etc.) to anchor boxes. Each entry contains a
            tensor  of shape
            `(H/stride * W/stride * num_anchors_per_location, 4)`,
            where H and W are the height and width of the image,
            stride is 2^level, and num_anchors_per_location is
            `num_scales * len(aspect_ratios)`.

    Example:
    ```python
    anchor_generator = AnchorGenerator(
        bounding_box_format='xyxy',
        min_level=3,
        max_level=7,
        num_scales=3,
        aspect_ratios=[0.5, 1.0, 2.0],
        anchor_size=4.0,
    )
    anchors = anchor_generator(images=keas.ops.ones(shape=(2, 640, 480, 3)))
    ```
    """

    def __init__(
        self,
        bounding_box_format,
        min_level,
        max_level,
        num_scales,
        aspect_ratios,
        anchor_size,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.bounding_box_format = bounding_box_format
        self.min_level = min_level
        self.max_level = max_level
        self.num_scales = num_scales
        self.aspect_ratios = aspect_ratios
        self.anchor_size = anchor_size
        self.built = True

    def call(self, inputs):
        images_shape = ops.shape(inputs)
        if len(images_shape) == 4:
            image_shape = images_shape[1:-1]
        else:
            image_shape = images_shape[:-1]

        image_shape = tuple(image_shape)

        multilevel_boxes = {}
        for level in range(self.min_level, self.max_level + 1):
            boxes_l = []
            # Calculate the feature map size for this level
            feat_size_y = math.ceil(image_shape[0] / 2**level)
            feat_size_x = math.ceil(image_shape[1] / 2**level)

            # Calculate the stride (step size) for this level
            stride_y = ops.cast(image_shape[0] / feat_size_y, "float32")
            stride_x = ops.cast(image_shape[1] / feat_size_x, "float32")

            # Generate anchor center points
            # Start from stride/2 to center anchors on pixels
            cx = ops.arange(stride_x / 2, image_shape[1], stride_x)
            cy = ops.arange(stride_y / 2, image_shape[0], stride_y)

            # Create a grid of anchor centers
            cx_grid, cy_grid = ops.meshgrid(cx, cy)

            for scale in range(self.num_scales):
                for aspect_ratio in self.aspect_ratios:
                    # Calculate the intermediate scale factor
                    intermidate_scale = 2 ** (scale / self.num_scales)
                    # Calculate the base anchor size for this level and scale
                    base_anchor_size = (
                        self.anchor_size * 2**level * intermidate_scale
                    )
                    # Adjust anchor dimensions based on aspect ratio
                    aspect_x = aspect_ratio**0.5
                    aspect_y = aspect_ratio**-0.5
                    half_anchor_size_x = base_anchor_size * aspect_x / 2.0
                    half_anchor_size_y = base_anchor_size * aspect_y / 2.0

                    # Generate anchor boxes (y1, x1, y2, x2 format)
                    boxes = ops.stack(
                        [
                            cy_grid - half_anchor_size_y,
                            cx_grid - half_anchor_size_x,
                            cy_grid + half_anchor_size_y,
                            cx_grid + half_anchor_size_x,
                        ],
                        axis=-1,
                    )
                    boxes_l.append(boxes)
            # Concat anchors on the same level to tensor shape HxWx(Ax4)
            boxes_l = ops.concatenate(boxes_l, axis=-1)
            boxes_l = ops.reshape(boxes_l, (-1, 4))
            # Convert to user defined
            multilevel_boxes[f"P{level}"] = convert_format(
                boxes_l,
                source="yxyx",
                target=self.bounding_box_format,
            )
        return multilevel_boxes

    def compute_output_shape(self, input_shape):
        multilevel_boxes_shape = {}
        if len(input_shape) == 4:
            image_height, image_width = input_shape[1:-1]
        else:
            image_height, image_width = input_shape[:-1]

        for i in range(self.min_level, self.max_level + 1):
            multilevel_boxes_shape[f"P{i}"] = (
                (image_height // 2 ** (i))
                * (image_width // 2 ** (i))
                * self.anchors_per_location,
                4,
            )
        return multilevel_boxes_shape

    @property
    def anchors_per_location(self):
        """
        The `anchors_per_location` property returns the number of anchors
        generated per pixel location, which is equal to
        `num_scales * len(aspect_ratios)`.
        """
        return self.num_scales * len(self.aspect_ratios)



from keras_hub.src.models.mit.mit_backbone import MiTBackbone
from keras_hub.src.models.mit.mit_image_classifier import MiTImageClassifier
from keras_hub.src.models.mit.mit_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, MiTBackbone)



"""vgg preset configurations."""

backbone_presets = {
    "vgg_11_imagenet": {
        "metadata": {
            "description": (
                "11-layer vgg model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 9220480,
            "official_name": "vgg",
            "path": "vgg",
            "model_card": "https://arxiv.org/abs/1409.1556",
        },
        "kaggle_handle": "kaggle://keras/vgg/keras/vgg_11_imagenet/1",
    },
    "vgg_13_imagenet": {
        "metadata": {
            "description": (
                "13-layer vgg model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 9404992,
            "official_name": "vgg",
            "path": "vgg",
            "model_card": "https://arxiv.org/abs/1409.1556",
        },
        "kaggle_handle": "kaggle://keras/vgg/keras/vgg_13_imagenet/1",
    },
    "vgg_16_imagenet": {
        "metadata": {
            "description": (
                "16-layer vgg model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 14714688,
            "official_name": "vgg",
            "path": "vgg",
            "model_card": "https://arxiv.org/abs/1409.1556",
        },
        "kaggle_handle": "kaggle://keras/vgg/keras/vgg_16_imagenet/1",
    },
    "vgg_19_imagenet": {
        "metadata": {
            "description": (
                "19-layer vgg model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 20024384,
            "official_name": "vgg",
            "path": "vgg",
            "model_card": "https://arxiv.org/abs/1409.1556",
        },
        "kaggle_handle": "kaggle://keras/vgg/keras/vgg_19_imagenet/1",
    },
}



import keras
from keras import ops


class BoxMatcher(keras.layers.Layer):
    """Box matching logic based on argmax of highest value (e.g., IOU).

    This class computes matches from a similarity matrix. Each row will be
    matched to at least one column, the matched result can either be positive
    or negative, or simply ignored depending on the setting.

    The settings include `thresholds` and `match_values`, for example if:
    1) `thresholds=[negative_threshold, positive_threshold]`, and
       `match_values=[negative_value=0, ignore_value=-1, positive_value=1]`: the
       rows will be assigned to positive_value if its argmax result >=
       positive_threshold; the rows will be assigned to negative_value if its
       argmax result < negative_threshold, and the rows will be assigned to
       ignore_value if its argmax result is between [negative_threshold,
       positive_threshold).
    2) `thresholds=[negative_threshold, positive_threshold]`, and
       `match_values=[ignore_value=-1, negative_value=0, positive_value=1]`: the
       rows will be assigned to positive_value if its argmax result >=
       positive_threshold; the rows will be assigned to ignore_value if its
       argmax result < negative_threshold, and the rows will be assigned to
       negative_value if its argmax result is between [negative_threshold,
       positive_threshold). This is different from case 1) by swapping first two
       values.
    3) `thresholds=[positive_threshold]`, and
       `match_values=[negative_values, positive_value]`: the rows will be
        assigned to positive value if its argmax result >= positive_threshold;
        the rows will be assigned to negative_value if its argmax result <
        negative_threshold.

    Args:
        thresholds: A sorted list of floats to classify the matches into
            different results (e.g. positive or negative or ignored match). The
            list will be prepended with -Inf and and appended with +Inf.
        match_values: A list of integers representing matched results (e.g.
            positive or negative or ignored match). len(`match_values`) must
            equal to len(`thresholds`) + 1.
        force_match_for_each_col: each row will be argmax matched to at
            least one column. This means some columns will be matched to
            multiple rows while some columns will not be matched to any rows.
            Filtering by `thresholds` will make less columns match to positive
            result. Setting this to True guarantees that each column will be
            matched to positive result to at least one row.

    Raises:
        ValueError: if `thresholds` not sorted or
        len(`match_values`) != len(`thresholds`) + 1

    Example:
    ```python
    box_matcher = keras_cv.layers.BoxMatcher([0.3, 0.7], [-1, 0, 1])
    iou_metric = keras_cv.bounding_box.compute_iou(anchors, boxes)
    matched_columns, matched_match_values = box_matcher(iou_metric)
    cls_mask = ops.less_equal(matched_match_values, 0)
    ```

    """

    def __init__(
        self,
        thresholds,
        match_values,
        force_match_for_each_col=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if sorted(thresholds) != thresholds:
            raise ValueError(f"`threshold` must be sorted, got {thresholds}")
        self.match_values = match_values
        if len(match_values) != len(thresholds) + 1:
            raise ValueError(
                f"len(`match_values`) must be len(`thresholds`) + 1, got "
                f"match_values {match_values}, thresholds {thresholds}"
            )
        thresholds.insert(0, -float("inf"))
        thresholds.append(float("inf"))
        self.thresholds = thresholds
        self.force_match_for_each_col = force_match_for_each_col
        self.built = True

    def call(self, similarity_matrix):
        """Matches each row to a column based on argmax

        Args:
            similarity_matrix: A float Tensor of shape `[num_rows, num_cols]` or
                `[batch_size, num_rows, num_cols]` representing any similarity
                metric.

        Returns:
            matched_columns: An integer tensor of shape `[num_rows]` or
                `[batch_size, num_rows]` storing the index of the matched
                column for each row.
            matched_values: An integer tensor of shape [num_rows] or
                `[batch_size, num_rows]` storing the match result
                `(positive match, negative match, ignored match)`.
        """
        squeeze_result = False
        if len(similarity_matrix.shape) == 2:
            squeeze_result = True
            similarity_matrix = ops.expand_dims(similarity_matrix, axis=0)
        static_shape = list(similarity_matrix.shape)
        num_rows = static_shape[1] or ops.shape(similarity_matrix)[1]
        batch_size = static_shape[0] or ops.shape(similarity_matrix)[0]

        def _match_when_cols_are_empty():
            """Performs matching when the rows of similarity matrix are empty.
            When the rows are empty, all detections are false positives. So we
            return a tensor of -1's to indicate that the rows do not match to
            any columns.

            Returns:
                matched_columns: An integer tensor of shape [batch_size,
                    num_rows] storing the index of the matched column for each
                    row.
                matched_values: An integer tensor of shape [batch_size,
                    num_rows] storing the match type indicator (e.g. positive or
                    negative or ignored match).
            """
            matched_columns = ops.zeros([batch_size, num_rows], dtype="int32")
            matched_values = -ops.ones([batch_size, num_rows], dtype="int32")
            return matched_columns, matched_values

        def _match_when_cols_are_non_empty():
            """Performs matching when the rows of similarity matrix are
            non-empty.
            Returns:
                matched_columns: An integer tensor of shape [batch_size,
                    num_rows] storing the index of the matched column for each
                    row.
                matched_values: An integer tensor of shape [batch_size,
                    num_rows] storing the match type indicator (e.g. positive or
                    negative or ignored match).
            """
            # Jax traces this function even when running eagerly and the
            # columns are non-empty. Therefore, we need to handle the case
            # where the similarity matrix is empty. We do this by padding
            # some -1s to the end. -1s are guaranteed to not affect argmax
            # matching because all values in a similarity matrix are [0,1]
            # and the indexing won't change because these are added at the
            # end.
            padded_similarity_matrix = ops.concatenate(
                [similarity_matrix, -ops.ones((batch_size, num_rows, 1))],
                axis=-1,
            )

            matched_columns = ops.argmax(
                padded_similarity_matrix,
                axis=-1,
            )

            # Get logical indices of ignored and unmatched columns as int32
            matched_vals = ops.max(padded_similarity_matrix, axis=-1)
            matched_values = ops.zeros([batch_size, num_rows], "int32")

            match_dtype = matched_vals.dtype
            for ind, low, high in zip(
                self.match_values, self.thresholds[:-1], self.thresholds[1:]
            ):
                low_threshold = ops.cast(low, match_dtype)
                high_threshold = ops.cast(high, match_dtype)
                mask = ops.logical_and(
                    ops.greater_equal(matched_vals, low_threshold),
                    ops.less(matched_vals, high_threshold),
                )
                matched_values = self._set_values_using_indicator(
                    matched_values, mask, ind
                )

            if self.force_match_for_each_col:
                # [batch_size, num_cols], for each column (groundtruth_box),
                # find the best matching row (anchor).
                matching_rows = ops.argmax(
                    padded_similarity_matrix,
                    axis=1,
                )
                # [batch_size, num_cols, num_rows], a transposed 0-1 mapping
                # matrix M, where M[j, i] = 1 means column j is matched to
                # row i.
                column_to_row_match_mapping = ops.one_hot(
                    matching_rows, num_rows
                )
                # [batch_size, num_rows], for each row (anchor), find the
                # matched column (groundtruth_box).
                force_matched_columns = ops.argmax(
                    column_to_row_match_mapping,
                    axis=1,
                )
                # [batch_size, num_rows]
                force_matched_column_mask = ops.cast(
                    ops.max(column_to_row_match_mapping, axis=1),
                    "bool",
                )
                # [batch_size, num_rows]
                matched_columns = ops.where(
                    force_matched_column_mask,
                    force_matched_columns,
                    matched_columns,
                )
                matched_values = ops.where(
                    force_matched_column_mask,
                    self.match_values[-1]
                    * ops.ones([batch_size, num_rows], dtype="int32"),
                    matched_values,
                )

            return ops.cast(matched_columns, "int32"), matched_values

        num_boxes = (
            similarity_matrix.shape[-1] or ops.shape(similarity_matrix)[-1]
        )
        matched_columns, matched_values = ops.cond(
            pred=ops.greater(num_boxes, 0),
            true_fn=_match_when_cols_are_non_empty,
            false_fn=_match_when_cols_are_empty,
        )

        if squeeze_result:
            matched_columns = ops.squeeze(matched_columns, axis=0)
            matched_values = ops.squeeze(matched_values, axis=0)

        return matched_columns, matched_values

    def _set_values_using_indicator(self, x, indicator, val):
        """Set the indicated fields of x to val.

        Args:
            x: tensor.
            indicator: boolean with same shape as x.
            val: scalar with value to set.
        Returns:
            modified tensor.
        """
        indicator = ops.cast(indicator, x.dtype)
        return ops.where(indicator == 0, x, val)

    def get_config(self):
        config = {
            "thresholds": self.thresholds[1:-1],
            "match_values": self.match_values,
            "force_match_for_each_col": self.force_match_for_each_col,
        }
        return config



import keras
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.vit_det.vit_layers import AddPositionalEmbedding
from keras_hub.src.models.vit_det.vit_layers import ViTDetPatchingAndEmbedding
from keras_hub.src.models.vit_det.vit_layers import WindowedTransformerEncoder


@keras_hub_export("keras_hub.models.ViTDetBackbone")
class ViTDetBackbone(Backbone):
    """An implementation of ViT image encoder.

    The ViTDetBackbone uses a windowed transformer encoder and relative
    positional encodings. The code has been adapted from [Segment Anything
    paper](https://arxiv.org/abs/2304.02643), [Segment Anything GitHub](
    https://github.com/facebookresearch/segment-anything) and [Detectron2](
    https://github.com/facebookresearch/detectron2).

    Args:
        hidden_size (int): The latent dimensionality to be projected
            into in the output of each stacked windowed transformer encoder.
        num_layers (int): The number of transformer encoder layers to
            stack in the Vision Transformer.
        intermediate_dim (int): The dimensionality of the hidden Dense
            layer in the transformer MLP head.
        num_heads (int): the number of heads to use in the
            `MultiHeadAttentionWithRelativePE` layer of each transformer
            encoder.
        global_attention_layer_indices (list): Indexes for blocks using
            global attention.
        image_shape (tuple[int], optional): The size of the input image in
            `(H, W, C)` format. Defaults to `(None, None, 3)`.
        patch_size (int, optional): the patch size to be supplied to the
            Patching layer to turn input images into a flattened sequence of
            patches. Defaults to `16`.
        num_output_channels (int, optional): The number of channels (features)
            in the output (image encodings). Defaults to `256`.
        use_bias (bool, optional): Whether to use bias to project the keys,
            queries, and values in the attention layer. Defaults to `True`.
        use_abs_pos (bool, optional): Whether to add absolute positional
            embeddings to the output patches. Defaults to `True`.
        use_rel_pos (bool, optional): Whether to use relative positional
            emcodings in the attention layer. Defaults to `True`.
        window_size (int, optional): The size of the window for windowed
            attention in the transformer encoder blocks. Defaults to `14`.
        layer_norm_epsilon (int, optional): The epsilon to use in the layer
            normalization blocks in transformer encoder. Defaults to `1e-6`.

    Examples:
    ```python
    input_data = np.ones((2, 224, 224, 3), dtype="float32")

    # Pretrained ViTDetBackbone backbone.
    model = keras_hub.models.ViTDetBackbone.from_preset("vit_det")
    model(input_data)

    # Randomly initialized ViTDetBackbone backbone with a custom config.
    model = keras_hub.models.ViTDetBackbone(
            image_shape = (16, 16, 3),
            patch_size = 2,
            hidden_size = 4,
            num_layers = 2,
            global_attention_layer_indices = [2, 5, 8, 11],
            intermediate_dim = 4 * 4,
            num_heads = 2,
            num_output_channels = 2,
            window_size = 2,
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        hidden_size,
        num_layers,
        intermediate_dim,
        num_heads,
        global_attention_layer_indices,
        image_shape=(None, None, 3),
        patch_size=16,
        num_output_channels=256,
        use_bias=True,
        use_abs_pos=True,
        use_rel_pos=True,
        window_size=14,
        layer_norm_epsilon=1e-6,
        **kwargs
    ):
        # === Functional model ===
        img_input = keras.layers.Input(shape=image_shape, name="images")
        # Check that the input image is well specified.
        if img_input.shape[-3] is None or img_input.shape[-2] is None:
            raise ValueError(
                "Height and width of the image must be specified"
                " in `image_shape`."
            )
        if img_input.shape[-3] != img_input.shape[-2]:
            raise ValueError(
                "Input image must be square i.e. the height must"
                " be equal to the width in the `image_shape`"
                " tuple/tensor."
            )
        img_size = img_input.shape[-3]
        x = img_input
        # VITDet scales inputs based on the standard ImageNet mean/stddev.
        x = (x - ops.array([0.485, 0.456, 0.406], dtype=x.dtype)) / (
            ops.array([0.229, 0.224, 0.225], dtype=x.dtype)
        )
        x = ViTDetPatchingAndEmbedding(
            kernel_size=(patch_size, patch_size),
            strides=(patch_size, patch_size),
            embed_dim=hidden_size,
        )(x)
        if use_abs_pos:
            x = AddPositionalEmbedding(img_size, patch_size, hidden_size)(x)
        for i in range(num_layers):
            x = WindowedTransformerEncoder(
                project_dim=hidden_size,
                intermediate_dim=intermediate_dim,
                num_heads=num_heads,
                use_bias=use_bias,
                use_rel_pos=use_rel_pos,
                window_size=(
                    window_size
                    if i not in global_attention_layer_indices
                    else 0
                ),
                input_size=(img_size // patch_size, img_size // patch_size),
            )(x)
        self.neck = keras.models.Sequential(
            [
                keras.layers.Conv2D(
                    filters=num_output_channels, kernel_size=1, use_bias=False
                ),
                keras.layers.LayerNormalization(epsilon=1e-6),
                keras.layers.Conv2D(
                    filters=num_output_channels,
                    kernel_size=3,
                    padding="same",
                    use_bias=False,
                ),
                keras.layers.LayerNormalization(epsilon=1e-6),
            ]
        )
        x = self.neck(x)

        super().__init__(inputs=img_input, outputs=x, **kwargs)

        # === Config ===
        self.patch_size = patch_size
        self.image_shape = image_shape
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.intermediate_dim = intermediate_dim
        self.num_heads = num_heads
        self.num_output_channels = num_output_channels
        self.use_bias = use_bias
        self.use_rel_pos = use_rel_pos
        self.use_abs_pos = use_abs_pos
        self.window_size = window_size
        self.global_attention_layer_indices = global_attention_layer_indices
        self.layer_norm_epsilon = layer_norm_epsilon

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_shape": self.image_shape,
                "patch_size": self.patch_size,
                "hidden_size": self.hidden_size,
                "num_layers": self.num_layers,
                "intermediate_dim": self.intermediate_dim,
                "num_heads": self.num_heads,
                "num_output_channels": self.num_output_channels,
                "use_bias": self.use_bias,
                "use_abs_pos": self.use_abs_pos,
                "use_rel_pos": self.use_rel_pos,
                "window_size": self.window_size,
                "global_attention_layer_indices": self.global_attention_layer_indices,
                "layer_norm_epsilon": self.layer_norm_epsilon,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.preprocessor import Preprocessor
from keras_hub.src.utils.tensor_utils import preprocessing_function


@keras_hub_export("keras_hub.models.ImageSegmenterPreprocessor")
class ImageSegmenterPreprocessor(Preprocessor):
    """Base class for image segmentation preprocessing layers.

    `ImageSegmenterPreprocessor` wraps a
    `keras_hub.layers.ImageConverter` to create a preprocessing layer for
    image segmentation tasks. It is intended to be paired with a
    `keras_hub.models.ImageSegmenter` task.

    All `ImageSegmenterPreprocessor` instances take three inputs: `x`, `y`, and
    `sample_weight`.

    - `x`: The first input, should always be included. It can be an image or
      a batch of images.
    - `y`: (Optional) Usually the segmentation mask(s), if `resize_output_mask`
        is set to `True` this will be resized to input image shape else will be
        passed through unaltered.
    - `sample_weight`: (Optional) Will be passed through unaltered.
    - `resize_output_mask` bool: If set to `True` the output mask will be resized to the same size as the input image. Defaults to `False`.

    The layer will output either `x`, an `(x, y)` tuple if labels were provided,
    or an `(x, y, sample_weight)` tuple if labels and sample weight were
    provided. `x` will be the input images after all model preprocessing has
    been applied.

    All `ImageSegmenterPreprocessor` tasks include a `from_preset()`
    constructor which can be used to load a pre-trained config.
    You can call the `from_preset()` constructor directly on this base class, in
    which case the correct class for your model will be automatically
    instantiated.

    Examples.
    ```python
    preprocessor = keras_hub.models.ImageSegmenterPreprocessor.from_preset(
        "deeplabv3_resnet50",
    )

    # Resize a single image for the model.
    x = np.ones((512, 512, 3))
    x = preprocessor(x)

    # Resize an image and its mask.
    x, y = np.ones((512, 512, 3)), np.zeros((512, 512, 1))
    x, y = preprocessor(x, y)

    # Resize a batch of images and masks.
    x, y = [np.ones((512, 512, 3)), np.zeros((512, 512, 3))],
           [np.ones((512, 512, 1)), np.zeros((512, 512, 1))]
    x, y = preprocessor(x, y)

    # Use a `tf.data.Dataset`.
    ds = tf.data.Dataset.from_tensor_slices((x, y)).batch(2)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)
    ```
    """

    def __init__(
        self,
        image_converter=None,
        resize_output_mask=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.image_converter = image_converter
        self.resize_output_mask = resize_output_mask

    @preprocessing_function
    def call(self, x, y=None, sample_weight=None):
        if self.image_converter:
            x = self.image_converter(x)

        if y is not None and self.image_converter and self.resize_output_mask:

            y = keras.layers.Resizing(
                height=(
                    self.image_converter.image_size[0]
                    if self.image_converter.image_size
                    else None
                ),
                width=(
                    self.image_converter.image_size[1]
                    if self.image_converter.image_size
                    else None
                ),
                crop_to_aspect_ratio=self.image_converter.crop_to_aspect_ratio,
                interpolation="nearest",
                data_format=self.image_converter.data_format,
                dtype=self.dtype_policy,
                name="mask_resizing",
            )(y)
        return keras.utils.pack_x_y_sample_weight(x, y, sample_weight)



import numpy as np

from keras_hub.src.models.pali_gemma.pali_gemma_backbone import (
    PaliGemmaBackbone,
)
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = PaliGemmaBackbone


def convert_backbone_config(transformers_config):
    text_config = transformers_config["text_config"]
    vision_config = transformers_config["vision_config"]
    return {
        "vocabulary_size": transformers_config["image_token_index"],
        "image_size": (
            vision_config["image_size"]
            if "image_size" in vision_config.keys()
            else 224
        ),
        "num_layers": text_config["num_hidden_layers"],
        "num_query_heads": text_config["num_attention_heads"],
        "num_key_value_heads": text_config["num_key_value_heads"],
        "hidden_dim": text_config["hidden_size"],
        "intermediate_dim": text_config["intermediate_size"] * 2,
        "head_dim": text_config["num_image_tokens"],
        "vit_patch_size": vision_config["patch_size"],
        "vit_num_heads": vision_config["num_attention_heads"],
        "vit_hidden_dim": vision_config["hidden_size"],
        "vit_num_layers": vision_config["num_hidden_layers"],
        "vit_intermediate_dim": vision_config["intermediate_size"],
    }


def convert_weights(backbone, loader, transformers_config):
    ############################################################################
    # Image Tower
    ############################################################################
    image_encoder = backbone.vit_encoder.get_layer("image_encoder")

    # Embedding
    loader.port_weight(
        keras_variable=image_encoder.vision_embeddings.patch_embedding.bias,
        hf_weight_key="vision_tower.vision_model.embeddings.patch_embedding.bias",
    )

    loader.port_weight(
        keras_variable=image_encoder.vision_embeddings.patch_embedding.kernel,
        hf_weight_key="vision_tower.vision_model.embeddings.patch_embedding.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(2, 3, 1, 0)),
    )

    # Positional Embedding
    loader.port_weight(
        keras_variable=image_encoder.vision_embeddings.position_embedding.embeddings,
        hf_weight_key="vision_tower.vision_model.embeddings.position_embedding.weight",
    )

    # Normalization
    loader.port_weight(
        keras_variable=image_encoder.encoder_layer_norm.gamma,
        hf_weight_key="vision_tower.vision_model.post_layernorm.weight",
    )

    loader.port_weight(
        keras_variable=image_encoder.encoder_layer_norm.beta,
        hf_weight_key="vision_tower.vision_model.post_layernorm.bias",
    )

    # ResBlocks
    for index in range(image_encoder.num_layers):
        block = image_encoder.resblocks[index]

        loader.port_weight(
            keras_variable=block.layer_norm_1.beta,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.layer_norm1.bias",
        )

        loader.port_weight(
            keras_variable=block.layer_norm_1.gamma,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.layer_norm1.weight",
        )

        loader.port_weight(
            keras_variable=block.layer_norm_2.beta,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.layer_norm2.bias",
        )

        loader.port_weight(
            keras_variable=block.layer_norm_2.gamma,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.layer_norm2.weight",
        )

        loader.port_weight(
            keras_variable=block.mlp_dense_1.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.mlp.fc1.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        loader.port_weight(
            keras_variable=block.mlp_dense_1.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.mlp.fc1.bias",
        )

        loader.port_weight(
            keras_variable=block.mlp_dense_2.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.mlp.fc2.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        loader.port_weight(
            keras_variable=block.mlp_dense_2.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.mlp.fc2.bias",
        )

        loader.port_weight(
            keras_variable=block.attn.key_proj.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.k_proj.bias",
        )

        loader.port_weight(
            keras_variable=block.attn.key_proj.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.k_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        loader.port_weight(
            keras_variable=block.attn.out_proj.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.out_proj.bias",
        )

        loader.port_weight(
            keras_variable=block.attn.out_proj.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.out_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        loader.port_weight(
            keras_variable=block.attn.query_proj.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.q_proj.bias",
        )

        loader.port_weight(
            keras_variable=block.attn.query_proj.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.q_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        loader.port_weight(
            keras_variable=block.attn.value_proj.bias,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.v_proj.bias",
        )

        loader.port_weight(
            keras_variable=block.attn.value_proj.kernel,
            hf_weight_key=f"vision_tower.vision_model.encoder.layers.{index}.self_attn.v_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

    # Multi Modal Projection
    loader.port_weight(
        keras_variable=backbone.vit_encoder.get_layer(
            "image_classifier"
        ).kernel,
        hf_weight_key="multi_modal_projector.linear.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
    )

    loader.port_weight(
        keras_variable=backbone.vit_encoder.get_layer("image_classifier").bias,
        hf_weight_key="multi_modal_projector.linear.bias",
    )

    ############################################################################
    # Language Tower
    ############################################################################
    for index in range(backbone.num_layers):
        decoder_layer = backbone.transformer_layers[index]

        # Norm layers
        loader.port_weight(
            keras_variable=decoder_layer.pre_attention_norm.scale,
            hf_weight_key=f"language_model.model.layers.{index}.input_layernorm.weight",
        )
        loader.port_weight(
            keras_variable=decoder_layer.pre_ffw_norm.scale,
            hf_weight_key=f"language_model.model.layers.{index}.post_attention_layernorm.weight",
        )

        # Attention layers
        loader.port_weight(
            keras_variable=decoder_layer.attention.query_dense.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.self_attn.q_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.key_dense.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.self_attn.k_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.value_dense.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.self_attn.v_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.output_dense.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.self_attn.o_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[2], keras_shape[0], keras_shape[1]),
                ),
                axes=(1, 2, 0),
            ),
        )

        # MLP layers
        loader.port_weight(
            keras_variable=decoder_layer.gating_ffw.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.mlp.gate_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.gating_ffw_2.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.mlp.up_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.ffw_linear.kernel,
            hf_weight_key=f"language_model.model.layers.{index}.mlp.down_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

    # Normalization
    loader.port_weight(
        keras_variable=backbone.layer_norm.scale,
        hf_weight_key="language_model.model.norm.weight",
    )

    # Embedding
    loader.port_weight(
        keras_variable=backbone.token_embedding.embeddings,
        hf_weight_key="language_model.model.embed_tokens.weight",
        hook_fn=lambda hf_tensor, keras_shape: hf_tensor[: keras_shape[0]],
    )


def convert_tokenizer(cls, preset, **kwargs):
    return cls(get_file(preset, "tokenizer.model"), **kwargs)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.gemma.gemma_backbone import GemmaBackbone
from keras_hub.src.models.gemma.gemma_tokenizer import GemmaTokenizer


@keras_hub_export("keras_hub.models.GemmaCausalLMPreprocessor")
class GemmaCausalLMPreprocessor(CausalLMPreprocessor):
    """Gemma Causal LM preprocessor.

    This preprocessing layer is meant for use with
    `keras_hub.models.GemmaCausalLM`. By default, it will take in batches of
    strings, and return outputs in a `(x, y, sample_weight)` format, where the
    `y` label is the next token id in the `x` sequence.

    For use with generation, the layer also exposes two methods
    `generate_preprocess()` and `generate_postprocess()`. When this preprocessor
    is attached to a `keras_hub.models.GemmaCausalLM` instance, these methods
    will be called implicitly in `generate()`. They can also be called
    standalone (e.g. to precompute preprocessing inputs for generation in a
    separate process).

    Args:
        tokenizer: A `keras_hub.models.GemmaTokenizer` instance.
        sequence_length: The length of the packed inputs.
        add_start_token: If `True`, the preprocessor will prepend the tokenizer
            start token to each input sequence.
        add_end_token: If `True`, the preprocessor will append the tokenizer
            end token to each input sequence.

    Call arguments:
        x: A string, `tf.Tensor` or list of python strings.
        y: Label data. Should always be `None` as the layer generates labels.
        sample_weight: Label weights. Should always be `None` as the layer
            generates label weights.
        sequence_length: Pass to override the configured `sequence_length` of
            the layer.

    Examples:
    ```python
    # Load the preprocessor from a preset.
    preprocessor = keras_hub.models.GemmaCausalLMPreprocessor.from_preset(
        "gemma_2b_en"
    )

    # Tokenize and pack a single sentence.
    preprocessor("The quick brown fox jumped.")

    # Tokenize a batch of sentences.
    preprocessor(["The quick brown fox jumped.", "Call me Ishmael."])

    # Apply tokenization to a `tf.data.Dataset`.
    features = tf.constant(["The quick brown fox.", "Call me Ishmael."])
    ds = tf.data.Dataset.from_tensor_slices(features)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Prepare tokens for generation (no end token).
    preprocessor.generate_preprocess(["The quick brown fox jumped."])

    # Map generation outputs back to strings.
    preprocessor.generate_postprocess({
        'token_ids': np.array([[2, 714, 4320, 8426, 25341, 32292, 235265, 0]]),
        'padding_mask': np.array([[ 1,  1,  1,  1,  1,  1,  1, 0]]),
    })
    ```
    """

    backbone_cls = GemmaBackbone
    tokenizer_cls = GemmaTokenizer



"""DistilBERT model preset configurations."""

backbone_presets = {
    "distil_bert_base_en_uncased": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where all input is lowercased. "
                "Trained on English Wikipedia + BooksCorpus using BERT as the "
                "teacher model."
            ),
            "params": 66362880,
            "official_name": "DistilBERT",
            "path": "distil_bert",
            "model_card": "https://huggingface.co/distilbert-base-uncased",
        },
        "kaggle_handle": "kaggle://keras/distil_bert/keras/distil_bert_base_en_uncased/2",
    },
    "distil_bert_base_en": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where case is maintained. "
                "Trained on English Wikipedia + BooksCorpus using BERT as the "
                "teacher model."
            ),
            "params": 65190912,
            "official_name": "DistilBERT",
            "path": "distil_bert",
            "model_card": "https://huggingface.co/distilbert-base-cased",
        },
        "kaggle_handle": "kaggle://keras/distil_bert/keras/distil_bert_base_en/2",
    },
    "distil_bert_base_multi": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where case is maintained. Trained on Wikipedias of 104 languages"
            ),
            "params": 134734080,
            "official_name": "DistilBERT",
            "path": "distil_bert",
            "model_card": "https://huggingface.co/distilbert-base-multilingual-cased",
        },
        "kaggle_handle": "kaggle://keras/distil_bert/keras/distil_bert_base_multi/2",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.albert.albert_backbone import AlbertBackbone
from keras_hub.src.models.albert.albert_tokenizer import AlbertTokenizer
from keras_hub.src.models.masked_lm_preprocessor import MaskedLMPreprocessor


@keras_hub_export("keras_hub.models.AlbertMaskedLMPreprocessor")
class AlbertMaskedLMPreprocessor(MaskedLMPreprocessor):
    """ALBERT preprocessing for the masked language modeling task.

    This preprocessing layer will prepare inputs for a masked language modeling
    task. It is primarily intended for use with the
    `keras_hub.models.AlbertMaskedLM` task model. Preprocessing will occur in
    multiple steps.

    - Tokenize any number of input segments using the `tokenizer`.
    - Pack the inputs together with the appropriate `"<s>"`, `"</s>"` and
      `"<pad>"` tokens, i.e., adding a single `"<s>"` at the start of the
      entire sequence, `"</s></s>"` between each segment,
      and a `"</s>"` at the end of the entire sequence.
    - Randomly select non-special tokens to mask, controlled by
      `mask_selection_rate`.
    - Construct a `(x, y, sample_weight)` tuple suitable for training with a
      `keras_hub.models.AlbertMaskedLM` task model.

    Args:
        tokenizer: A `keras_hub.models.AlbertTokenizer` instance.
        sequence_length: The length of the packed inputs.
        mask_selection_rate: The probability an input token will be dynamically
            masked.
        mask_selection_length: The maximum number of masked tokens supported
            by the layer.
        mask_token_rate: float. `mask_token_rate` must be
            between 0 and 1 which indicates how often the mask_token is
            substituted for tokens selected for masking.
            Defaults to `0.8`.
        random_token_rate: float. `random_token_rate` must be
            between 0 and 1 which indicates how often a random token is
            substituted for tokens selected for masking. Default is 0.1.
            Note: mask_token_rate + random_token_rate <= 1,  and for
            (1 - mask_token_rate - random_token_rate), the token will not be
            changed. Defaults to `0.1`.
        truncate: string. The algorithm to truncate a list of batched segments
            to fit within `sequence_length`. The value can be either
            `round_robin` or `waterfall`:
            - `"round_robin"`: Available space is assigned one token at a
                time in a round-robin fashion to the inputs that still need
                some, until the limit is reached.
            - `"waterfall"`: The allocation of the budget is done using a
                "waterfall" algorithm that allocates quota in a
                left-to-right manner and fills up the buckets until we run
                out of budget. It supports an arbitrary number of segments.

    Examples:

    Directly calling the layer on data.
    ```python
    preprocessor = keras_hub.models.AlbertMaskedLMPreprocessor.from_preset(
        "albert_base_en_uncased"
    )

    # Tokenize and mask a single sentence.
    preprocessor("The quick brown fox jumped.")

    # Tokenize and mask a batch of single sentences.
    preprocessor(["The quick brown fox jumped.", "Call me Ishmael."])

    # Tokenize and mask sentence pairs.
    # In this case, always convert input to tensors before calling the layer.
    first = tf.constant(["The quick brown fox jumped.", "Call me Ishmael."])
    second = tf.constant(["The fox tripped.", "Oh look, a whale."])
    preprocessor((first, second))
    ```

    Mapping with `tf.data.Dataset`.
    ```python
    preprocessor = keras_hub.models.AlbertMaskedLMPreprocessor.from_preset(
        "albert_base_en_uncased"
    )

    first = tf.constant(["The quick brown fox jumped.", "Call me Ishmael."])
    second = tf.constant(["The fox tripped.", "Oh look, a whale."])

    # Map single sentences.
    ds = tf.data.Dataset.from_tensor_slices(first)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Map sentence pairs.
    ds = tf.data.Dataset.from_tensor_slices((first, second))
    # Watch out for tf.data's default unpacking of tuples here!
    # Best to invoke the `preprocessor` directly in this case.
    ds = ds.map(
        lambda first, second: preprocessor(x=(first, second)),
        num_parallel_calls=tf.data.AUTOTUNE,
    )
    ```
    """

    backbone_cls = AlbertBackbone
    tokenizer_cls = AlbertTokenizer



"""Falcon model preset configurations."""

backbone_presets = {
    "falcon_refinedweb_1b_en": {
        "metadata": {
            "description": (
                "24-layer Falcon model (Falcon with 1B parameters), trained on "
                "350B tokens of RefinedWeb dataset."
            ),
            "params": 1311625216,
            "official_name": "Falcon",
            "path": "falcon",
            "model_card": "https://huggingface.co/tiiuae/falcon-rw-1b",
        },
        "kaggle_handle": "kaggle://keras/falcon/keras/falcon_refinedweb_1b_en/1",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.densenet.densenet_backbone import DenseNetBackbone


@keras_hub_export("keras_hub.layers.DenseNetImageConverter")
class DenseNetImageConverter(ImageConverter):
    backbone_cls = DenseNetBackbone



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.f_net.f_net_backbone import FNetBackbone
from keras_hub.src.models.f_net.f_net_tokenizer import FNetTokenizer
from keras_hub.src.models.text_classifier_preprocessor import (
    TextClassifierPreprocessor,
)
from keras_hub.src.utils.tensor_utils import preprocessing_function


@keras_hub_export(
    [
        "keras_hub.models.FNetTextClassifierPreprocessor",
        "keras_hub.models.FNetPreprocessor",
    ]
)
class FNetTextClassifierPreprocessor(TextClassifierPreprocessor):
    """An FNet preprocessing layer which tokenizes and packs inputs.

    This preprocessing layer will do three things:

     1. Tokenize any number of input segments using the `tokenizer`.
     2. Pack the inputs together using a `keras_hub.layers.MultiSegmentPacker`.
       with the appropriate `"[CLS]"`, `"[SEP]"` and `"<pad>"` tokens.
     3. Construct a dictionary with keys `"token_ids"`, and `"segment_ids"`  that
       can be passed directly to `keras_hub.models.FNetBackbone`.

    This layer can be used directly with `tf.data.Dataset.map` to preprocess
    string data in the `(x, y, sample_weight)` format used by
    `keras.Model.fit`.

    Args:
        tokenizer: A `keras_hub.models.FNetTokenizer` instance.
        sequence_length: The length of the packed inputs.
        truncate: string. The algorithm to truncate a list of batched segments
            to fit within `sequence_length`. The value can be either
            `round_robin` or `waterfall`:
            - `"round_robin"`: Available space is assigned one token at a
                time in a round-robin fashion to the inputs that still need
                some, until the limit is reached.
            - `"waterfall"`: The allocation of the budget is done using a
                "waterfall" algorithm that allocates quota in a
                left-to-right manner and fills up the buckets until we run
                out of budget. It supports an arbitrary number of segments.

    Call arguments:
        x: A tensor of single string sequences, or a tuple of multiple
            tensor sequences to be packed together. Inputs may be batched or
            unbatched. For single sequences, raw python inputs will be converted
            to tensors. For multiple sequences, pass tensors directly.
        y: Any label data. Will be passed through unaltered.
        sample_weight: Any label weight data. Will be passed through unaltered.

    Examples:

    Directly calling the from_preset().
    ```python
    preprocessor = keras_hub.models.TextClassifierPreprocessor.from_preset(
        "f_net_base_en"
    )

    # Tokenize and pack a single sentence.
    preprocessor("The quick brown fox jumped.")

    # Tokenize and a batch of single sentences.
    preprocessor(["The quick brown fox jumped.", "Call me Ishmael."])

    # Preprocess a batch of sentence pairs.
    # When handling multiple sequences, always convert to tensors first!
    first = tf.constant(["The quick brown fox jumped.", "Call me Ishmael."])
    second = tf.constant(["The fox tripped.", "Oh look, a whale."])
    preprocessor((first, second))
    ```

    Mapping with `tf.data.Dataset`.
    ```python
    preprocessor = keras_hub.models.TextClassifierPreprocessor.from_preset(
        "f_net_base_en"
    )
    first = tf.constant(["The quick brown fox jumped.", "Call me Ishmael."])
    second = tf.constant(["The fox tripped.", "Oh look, a whale."])
    label = tf.constant([1, 1])

    # Map labeled single sentences.
    ds = tf.data.Dataset.from_tensor_slices((first, label))
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Map unlabeled single sentences.
    ds = tf.data.Dataset.from_tensor_slices(first)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Map labeled sentence pairs.
    ds = tf.data.Dataset.from_tensor_slices(((first, second), label))
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Map unlabeled sentence pairs.
    ds = tf.data.Dataset.from_tensor_slices((first, second))

    # Watch out for tf.data's default unpacking of tuples here!
    # Best to invoke the `preprocessor` directly in this case.
    ds = ds.map(
        lambda first, second: preprocessor(x=(first, second)),
        num_parallel_calls=tf.data.AUTOTUNE,
    )
    ```
    """

    backbone_cls = FNetBackbone
    tokenizer_cls = FNetTokenizer

    @preprocessing_function
    def call(self, x, y=None, sample_weight=None):
        # FNet has not padding mask.
        output = super().call(x, y=y, sample_weight=sample_weight)
        x, y, sample_weight = keras.utils.unpack_x_y_sample_weight(output)
        del x["padding_mask"]
        return keras.utils.pack_x_y_sample_weight(x, y, sample_weight)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.tokenizers.byte_pair_tokenizer import BytePairTokenizer
from keras_hub.src.tokenizers.byte_pair_tokenizer import convert_to_ragged_batch
from keras_hub.src.tokenizers.byte_pair_tokenizer import split_strings_for_bpe
from keras_hub.src.utils.tensor_utils import preprocessing_function

try:
    import tensorflow as tf
except ImportError:
    tf = None


@keras_hub_export(
    [
        "keras_hub.tokenizers.CLIPTokenizer",
        "keras_hub.models.CLIPTokenizer",
    ]
)
class CLIPTokenizer(BytePairTokenizer):
    """A CLIP tokenizer using Byte-Pair Encoding subword segmentation.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.BytePairTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by CLIP
    models and provides a `from_preset()` method to automatically download
    a matching vocabulary for a CLIP preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        vocabulary: string or dict, maps token to integer ids. If it is a
            string, it should be the file path to a json file.
        merges: string or list, contains the merge rule. If it is a string,
            it should be the file path to merge rules. The merge rule file
            should have one merge rule per line. Every merge rule contains
            merge entities separated by a space.
        pad_with_end_token: bool. Whether to pad the output with `end_token`.
    """

    # TODO: Add example and `backbone_cls` once we have a CLIP model.

    backbone_cls = None

    def __init__(
        self,
        vocabulary=None,
        merges=None,
        pad_with_end_token=False,
        **kwargs,
    ):
        self._add_special_token("<|startoftext|>", "start_token")
        self._add_special_token("<|endoftext|>", "end_token")
        self.pad_token_id = 0
        self.pad_with_end_token = pad_with_end_token

        super().__init__(
            vocabulary=vocabulary,
            merges=merges,
            unsplittable_tokens=[self.start_token, self.end_token],
            **kwargs,
        )

    def set_vocabulary_and_merges(self, vocabulary, merges):
        super().set_vocabulary_and_merges(vocabulary, merges)
        if self.pad_with_end_token:
            self.pad_token_id = self.end_token_id

    def _bpe_merge_and_update_cache(self, tokens):
        """Process unseen tokens and add to cache."""
        words = self._transform_bytes(tokens)

        # In CLIP, we need to add `</w>` to the last word.
        words = tf.strings.reduce_join(words, axis=1, separator=" ")
        words = tf.strings.join([words, "</w>"])
        words = tf.strings.split(words, sep=" ")
        tokenized_words = self._bpe_merge(words)

        # For each word, join all its token by a whitespace,
        # e.g., ["dragon", "fly"] => "dragon fly" for hash purpose.
        tokenized_words = tf.strings.reduce_join(
            tokenized_words, axis=1, separator=" "
        )
        self.cache.insert(tokens, tokenized_words)

    @preprocessing_function
    def tokenize(self, inputs):
        self._check_vocabulary()
        if self.add_prefix_space:
            inputs = tf.strings.join([" ", inputs])

        unbatched = inputs.shape.rank == 0
        if unbatched:
            inputs = tf.expand_dims(inputs, 0)
        if inputs.shape.rank > 1:
            raise ValueError(
                "`tokenize()` inputs should be a string, list of strings, or "
                f"string tensor with rank < 2. Received: {inputs}"
            )

        raw_tokens = split_strings_for_bpe(inputs, self.unsplittable_tokens)

        # Strip and remove empty tokens.
        raw_tokens = tf.strings.strip(raw_tokens)
        raw_tokens = tf.ragged.boolean_mask(raw_tokens, raw_tokens != "")

        token_row_splits = raw_tokens.row_splits
        flat_tokens = raw_tokens.flat_values

        # Check cache.
        cache_lookup = self.cache.lookup(flat_tokens)
        cache_mask = cache_lookup == ""

        has_unseen_words = tf.math.reduce_any(
            (cache_lookup == "") & (flat_tokens != "")
        )

        def process_unseen_tokens():
            unseen_tokens = tf.boolean_mask(flat_tokens, cache_mask)
            self._bpe_merge_and_update_cache(unseen_tokens)
            return self.cache.lookup(flat_tokens)

        # If `has_unseen_words == True`, it means not all tokens are in cache,
        # we will process the unseen tokens. Otherwise return the cache lookup.
        tokenized_words = tf.cond(
            has_unseen_words,
            process_unseen_tokens,
            lambda: cache_lookup,
        )

        tokens = tf.strings.split(tokenized_words, sep=" ")
        if self.compute_dtype != tf.string:
            # Encode merged tokens.
            tokens = self.token_to_id_map.lookup(tokens)

        # Unflatten to match input.
        tokens = tf.RaggedTensor.from_row_splits(
            tokens.flat_values,
            tf.gather(tokens.row_splits, token_row_splits),
        )

        # Convert to a dense output if `sequence_length` is set.
        if self.sequence_length:
            output_shape = tokens.shape.as_list()
            output_shape[-1] = self.sequence_length
            tokens = tokens.to_tensor(shape=output_shape)

        # Convert to a dense output if input in scalar
        if unbatched:
            tokens = tf.squeeze(tokens, 0)
            tf.ensure_shape(tokens, shape=[self.sequence_length])

        return tokens

    @preprocessing_function
    def detokenize(self, inputs):
        self._check_vocabulary()
        inputs, unbatched, _ = convert_to_ragged_batch(inputs)
        inputs = tf.cast(inputs, self.dtype)
        unicode_text = tf.strings.reduce_join(
            self.id_to_token_map.lookup(inputs), axis=-1
        )

        # When detokenizing, we need to remove </w> and extra whitespace.
        unicode_text = tf.strings.regex_replace(unicode_text, r"</w>", " ")
        unicode_text = tf.strings.strip(unicode_text)

        split_unicode_text = tf.strings.unicode_split(unicode_text, "UTF-8")
        outputs = tf.strings.reduce_join(
            self.unicode2byte.lookup(split_unicode_text), axis=-1
        )

        if unbatched:
            outputs = tf.squeeze(outputs, 0)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "pad_with_end_token": self.pad_with_end_token,
            }
        )
        # In the constructor, we pass the list of special tokens to the
        # `unsplittable_tokens` arg of the superclass' constructor. Hence, we
        # delete it from the config here.
        del config["unsplittable_tokens"]
        return config



from keras_hub.src.models.opt.opt_backbone import OPTBackbone
from keras_hub.src.models.opt.opt_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, OPTBackbone)



import numpy as np
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_segmenter import ImageSegmenter
from keras_hub.src.models.sam.sam_backbone import SAMBackbone
from keras_hub.src.models.sam.sam_image_segmenter_preprocessor import (
    SAMImageSegmenterPreprocessor,
)


@keras_hub_export("keras_hub.models.SAMImageSegmenter")
class SAMImageSegmenter(ImageSegmenter):
    """The Segment Anything (SAM) image segmenter Model.

    SAM works by prompting the input images. There are three ways to prompt:
    (1) Labelled Points: Foreground points (points with label 1) are encoded
        such that the output masks generated by the mask decoder contain them
        and background points (points with label 0) are encoded such that the
        generated masks don't contain them.
    (2) Box: A box tells the model which part/crop of the image to segment.
    (3) Mask: An input mask can be used to refine the output of the mask
        decoder.
    These prompts can be mixed and matched but at least one of the prompts
    must be present. To turn off a particular prompt, simply exclude it from
    the inputs to the model.
    (1) For points prompts, the expected shape is `(batch, num_points, 2)`.
        The labels must have a corresponding shape of `(batch, num_points)`.
    (2) For box prompt, the expected shape is `(batch, 1, 2, 2)`.
    (3) Similarly, mask prompts have shape `(batch, 1, H, W, 1)`.


    Args:
      backbone: A `keras_hub.models.SAMBackbone` instance.

    Example:
    Load pretrained model using `from_preset`.

    ```python
    image_size=128
    batch_size=2
    input_data = {
        "images": np.ones(
            (batch_size, image_size, image_size, 3),
            dtype="float32",
        ),
        "points": np.ones((batch_size, 1, 2), dtype="float32"),
        "labels": np.ones((batch_size, 1), dtype="float32"),
        "boxes": np.ones((batch_size, 1, 2, 2), dtype="float32"),
        "masks": np.zeros(
            (batch_size, 0, image_size, image_size, 1)
        ),
    }
    # todo: update preset name
    sam = keras_hub.models.SAMImageSegmenter.from_preset(`sam_base`)
    sam(input_data)
    ```

    Load segment anything image segmenter with custom backbone

    ```python
    image_size = 128
    batch_size = 2
    images = np.ones(
        (batch_size, image_size, image_size, 3),
        dtype="float32",
    )
    image_encoder = ViTDetBackbone(
        hidden_size=16,
        num_layers=16,
        intermediate_dim=16 * 4,
        num_heads=16,
        global_attention_layer_indices=[2, 5, 8, 11],
        patch_size=16,
        num_output_channels=8,
        window_size=2,
        image_shape=(image_size, image_size, 3),
    )
    prompt_encoder = SAMPromptEncoder(
        hidden_size=8,
        image_embedding_size=(8, 8),
        input_image_size=(
            image_size,
            image_size,
        ),
        mask_in_channels=16,
    )
    mask_decoder = SAMMaskDecoder(
        num_layers=2,
        hidden_size=8,
        intermediate_dim=32,
        num_heads=8,
        embedding_dim=8,
        num_multimask_outputs=3,
        iou_head_depth=3,
        iou_head_hidden_dim=8,
    )
    backbone = SAMBackbone(
        image_encoder=image_encoder,
        prompt_encoder=prompt_encoder,
        mask_decoder=mask_decoder,
        image_shape=(image_size, image_size, 3),
    )
    sam = SAMImageSegmenter(
        backbone=backbone
    )
    ```

    For example, to pass in all the prompts, do:

    ```python

    points = np.array([[[512., 512.], [100., 100.]]])
    # For labels: 1 means foreground point, 0 means background
    labels = np.array([[1., 0.]])
    box = np.array([[[[384., 384.], [640., 640.]]]])
    input_mask = np.ones((1, 1, 256, 256, 1))
    Prepare an input dictionary:
    inputs = {
        "images": image,
        "points": points,
        "labels": labels,
        "boxes": box,
        "masks": input_mask
    }
    outputs = sam.predict(inputs)
    masks, iou_pred = outputs["masks"], outputs["iou_pred"]
    ```

    The first mask in the output `masks` (i.e. `masks[:, 0, ...]`) is the best
    mask predicted by the model based on the prompts. Other `masks`
    (i.e. `masks[:, 1:, ...]`) are alternate predictions that can be used if
    they are desired over the first one.
    Now, in case of only points and box prompts, simply exclude the masks:

    ```python
    inputs = {
        "images": image,
        "points": points,
        "labels": labels,
        "boxes": box,
    }

    outputs = sam.predict(inputs)
    masks, iou_pred = outputs["masks"], outputs["iou_pred"]
    ```

    Another example is that only points prompts are present.
    Note that if point prompts are present but no box prompt is present, the
    points must be padded using a zero point and -1 label:

    ```python
    padded_points = np.concatenate(
        [points, np.zeros((1, 1, 2))], axis=1
    )

    padded_labels = np.concatenate(
        [labels, -np.ones((1, 1))], axis=1
    )
    inputs = {
        "images": image,
        "points": padded_points,
        "labels": padded_labels,
    }
    outputs = sam.predict(inputs)
    masks, iou_pred = outputs["masks"], outputs["iou_pred"]
    ```
    """

    backbone_cls = SAMBackbone
    preprocessor_cls = SAMImageSegmenterPreprocessor

    def __init__(self, backbone, preprocessor=None, **kwargs):
        # The implementation has been adapted form [Segment Anything
        # paper](https://arxiv.org/abs/2304.02643) and [Segment Anything
        # GitHub](https://github.com/facebookresearch/segment-anything) and
        # [Detectron2](https://github.com/facebookresearch/detectron2).
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        # === Functional Model ===
        inputs = self.backbone.input
        x = self.backbone(inputs)
        outputs = self.backbone.mask_decoder(**x)
        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

    def predict_step(self, *args, **kwargs):
        if len(args) == 2:
            args = (args[0], self._add_placeholder_prompts(args[-1]))
        else:
            args = (self._add_placeholder_prompts(args[0]),)

        return super().predict_step(*args, **kwargs)

    def fit(self, *args, **kwargs):
        raise NotImplementedError(
            "Segment Anything Model only supports inference for now. Training"
            " the model isn't supported yet."
        )

    def _add_placeholder_prompts(self, inputs):
        """Adds placeholder prompt inputs for a call to SAM.

        Because SAM is a functional subclass model, all inputs must be specified in
        calls to the model. However, prompt inputs are all optional, so we have to
        add placeholders when they're not specified by the user.
        """
        inputs = inputs.copy()

        # Get the batch shape based on the image input
        batch_size = ops.shape(inputs["images"])[0]

        # The type of the placeholders must match the existing inputs with respect
        # to whether or not they are tensors (as opposed to Numpy arrays).
        zeros = ops.zeros if ops.is_tensor(inputs["images"]) else np.zeros

        # Fill in missing inputs.
        if "points" not in inputs:
            inputs["points"] = zeros((batch_size, 0, 2))
        if "labels" not in inputs:
            inputs["labels"] = zeros((batch_size, 0))
        if "boxes" not in inputs:
            inputs["boxes"] = zeros((batch_size, 0, 2, 2))
        if "masks" not in inputs:
            inputs["masks"] = zeros((batch_size, 0, 256, 256, 1))

        return inputs



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone

BN_EPSILON = 1.001e-5


@keras_hub_export("keras_hub.models.DenseNetBackbone")
class DenseNetBackbone(FeaturePyramidBackbone):
    """Instantiates the DenseNet architecture.

    This class implements a DenseNet backbone as described in
    [Densely Connected Convolutional Networks (CVPR 2017)](
       https://arxiv.org/abs/1608.06993
    ).

    Args:
        stackwise_num_repeats: list of ints, number of repeated convolutional
            blocks per dense block.
        image_shape: optional shape tuple, defaults to (None, None, 3).
        compression_ratio: float, compression rate at transition layers,
            defaults to 0.5.
        growth_rate: int, number of filters added by each dense block,
            defaults to 32

    Examples:
    ```python
    input_data = np.ones(shape=(8, 224, 224, 3))

    # Pretrained backbone
    model = keras_hub.models.DenseNetBackbone.from_preset("densenet121_imagenet")
    model(input_data)

    # Randomly initialized backbone with a custom config
    model = keras_hub.models.DenseNetBackbone(
        stackwise_num_repeats=[6, 12, 24, 16],
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        stackwise_num_repeats,
        image_shape=(None, None, 3),
        compression_ratio=0.5,
        growth_rate=32,
        **kwargs,
    ):
        # === Functional Model ===
        data_format = keras.config.image_data_format()
        channel_axis = -1 if data_format == "channels_last" else 1
        image_input = keras.layers.Input(shape=image_shape)

        x = image_input  # Intermediate result.
        x = keras.layers.Conv2D(
            64,
            7,
            strides=2,
            use_bias=False,
            padding="same",
            data_format=data_format,
            name="conv1_conv",
        )(x)
        x = keras.layers.BatchNormalization(
            axis=channel_axis, epsilon=BN_EPSILON, name="conv1_bn"
        )(x)
        x = keras.layers.Activation("relu", name="conv1_relu")(x)
        x = keras.layers.MaxPooling2D(
            3, strides=2, padding="same", data_format=data_format, name="pool1"
        )(x)

        pyramid_outputs = {}
        for stack_index in range(len(stackwise_num_repeats) - 1):
            index = stack_index + 2
            x = apply_dense_block(
                x,
                channel_axis,
                stackwise_num_repeats[stack_index],
                growth_rate,
                name=f"stack{stack_index+1}",
            )
            pyramid_outputs[f"P{index}"] = x
            x = apply_transition_block(
                x,
                channel_axis,
                compression_ratio,
                name=f"transition{stack_index+1}",
            )

        x = apply_dense_block(
            x,
            channel_axis,
            stackwise_num_repeats[-1],
            growth_rate,
            name=f"stack{len(stackwise_num_repeats)}",
        )
        pyramid_outputs[f"P{len(stackwise_num_repeats) + 1}"] = x
        x = keras.layers.BatchNormalization(
            axis=channel_axis, epsilon=BN_EPSILON, name="bn"
        )(x)
        x = keras.layers.Activation("relu", name="relu")(x)

        super().__init__(inputs=image_input, outputs=x, **kwargs)

        # === Config ===
        self.stackwise_num_repeats = stackwise_num_repeats
        self.compression_ratio = compression_ratio
        self.growth_rate = growth_rate
        self.image_shape = image_shape
        self.pyramid_outputs = pyramid_outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "stackwise_num_repeats": self.stackwise_num_repeats,
                "compression_ratio": self.compression_ratio,
                "growth_rate": self.growth_rate,
                "image_shape": self.image_shape,
            }
        )
        return config


def apply_dense_block(x, channel_axis, num_repeats, growth_rate, name=None):
    """A dense block.

    Args:
      x: input tensor.
      num_repeats: int, number of repeated convolutional blocks.
      growth_rate: int, number of filters added by each dense block.
      name: string, block label.
    """
    if name is None:
        name = f"dense_block_{keras.backend.get_uid('dense_block')}"

    for i in range(num_repeats):
        x = apply_conv_block(
            x, channel_axis, growth_rate, name=f"{name}_block{i+1}"
        )
    return x


def apply_transition_block(x, channel_axis, compression_ratio, name=None):
    """A transition block.

    Args:
      x: input tensor.
      compression_ratio: float, compression rate at transition layers.
      name: string, block label.
    """
    data_format = keras.config.image_data_format()
    if name is None:
        name = f"transition_block_{keras.backend.get_uid('transition_block')}"

    x = keras.layers.BatchNormalization(
        axis=channel_axis, epsilon=BN_EPSILON, name=f"{name}_bn"
    )(x)
    x = keras.layers.Activation("relu", name=f"{name}_relu")(x)
    x = keras.layers.Conv2D(
        int(x.shape[channel_axis] * compression_ratio),
        1,
        use_bias=False,
        data_format=data_format,
        name=f"{name}_conv",
    )(x)
    x = keras.layers.AveragePooling2D(
        2, strides=2, data_format=data_format, name=f"{name}_pool"
    )(x)
    return x


def apply_conv_block(x, channel_axis, growth_rate, name=None):
    """A building block for a dense block.

    Args:
      x: input tensor.
      growth_rate: int, number of filters added by each dense block.
      name: string, block label.
    """
    data_format = keras.config.image_data_format()
    if name is None:
        name = f"conv_block_{keras.backend.get_uid('conv_block')}"

    shortcut = x
    x = keras.layers.BatchNormalization(
        axis=channel_axis, epsilon=BN_EPSILON, name=f"{name}_1_bn"
    )(x)
    x = keras.layers.Activation("relu", name=f"{name}_1_relu")(x)
    x = keras.layers.Conv2D(
        4 * growth_rate,
        1,
        use_bias=False,
        data_format=data_format,
        name=f"{name}_1_conv",
    )(x)
    x = keras.layers.BatchNormalization(
        axis=channel_axis, epsilon=BN_EPSILON, name=f"{name}_2_bn"
    )(x)
    x = keras.layers.Activation("relu", name=f"{name}_2_relu")(x)
    x = keras.layers.Conv2D(
        growth_rate,
        3,
        padding="same",
        use_bias=False,
        data_format=data_format,
        name=f"{name}_2_conv",
    )(x)
    x = keras.layers.Concatenate(axis=channel_axis, name=f"{name}_concat")(
        [shortcut, x]
    )
    return x



"""XLM-RoBERTa model preset configurations."""

backbone_presets = {
    "t5_small_multi": {
        "metadata": {
            "description": (
                "8-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/t5_small_multi/2",
    },
    "t5_base_multi": {
        "metadata": {
            "description": (
                "12-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/t5_base_multi/2",
    },
    "t5_large_multi": {
        "metadata": {
            "description": (
                "24-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/t5_large_multi/2",
    },
    "flan_small_multi": {
        "metadata": {
            "description": (
                "8-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/flan_small_multi/2",
    },
    "flan_base_multi": {
        "metadata": {
            "description": (
                "12-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/flan_base_multi/2",
    },
    "flan_large_multi": {
        "metadata": {
            "description": (
                "24-layer T5 model. Trained on the Colossal Clean Crawled "
                "Corpus (C4)."
            ),
            "params": 0,
            "official_name": "T5",
            "path": "t5",
            "model_card": "https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.md",
        },
        "kaggle_handle": "kaggle://keras/t5/keras/flan_large_multi/2",
    },
}



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.task import Task


@keras_hub_export("keras_hub.models.ImageSegmenter")
class ImageSegmenter(Task):
    """Base class for all image segmentation tasks.

    `ImageSegmenter` tasks wrap a `keras_hub.models.Task` and
    a `keras_hub.models.Preprocessor` to create a model that can be used for
    image segmentation.

    All `ImageSegmenter` tasks include a `from_preset()` constructor which can
    be used to load a pre-trained config and weights.
    """

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        metrics="auto",
        **kwargs,
    ):
        """Configures the `ImageSegmenter` task for training.

        The `ImageSegmenter` task extends the default compilation signature of
        `keras.Model.compile` with defaults for `optimizer`, `loss`, and
        `metrics`. To override these defaults, pass any value
        to these arguments during compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default optimizer
                for the given model and task. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer` values.
            loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, where a
                `keras.losses.SparseCategoricalCrossentropy` loss will be
                applied for the classification task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `"auto"`, or a list of metrics to be evaluated by
                the model during training and testing. Defaults to `"auto"`,
                where a `keras.metrics.SparseCategoricalAccuracy` will be
                applied to track the accuracy of the model during training.
                See `keras.Model.compile` and `keras.metrics` for
                more info on possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if optimizer == "auto":
            optimizer = keras.optimizers.Adam(5e-5)
        if loss == "auto":
            activation = getattr(self, "activation", None)
            activation = keras.activations.get(activation)
            from_logits = activation != keras.activations.softmax
            loss = keras.losses.CategoricalCrossentropy(from_logits=from_logits)
        if metrics == "auto":
            metrics = [keras.metrics.CategoricalAccuracy()]
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            **kwargs,
        )



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.start_end_packer import StartEndPacker
from keras_hub.src.models.preprocessor import Preprocessor
from keras_hub.src.models.t5.t5_tokenizer import T5Tokenizer
from keras_hub.src.utils.tensor_utils import preprocessing_function


@keras_hub_export("keras_hub.models.T5Preprocessor")
class T5Preprocessor(Preprocessor):
    tokenizer_cls = T5Tokenizer

    def __init__(
        self,
        tokenizer,
        sequence_length=256,
        add_start_token=False,
        add_end_token=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.tokenizer = tokenizer
        self.sequence_length = sequence_length
        self.add_start_token = add_start_token
        self.add_end_token = add_end_token

    def build(self, input_shape):
        # Defer packer creation to `build()` so that we can be sure tokenizer
        # assets have loaded when restoring a saved model.
        self.packer = StartEndPacker(
            start_value=self.tokenizer.start_token_id,
            end_value=self.tokenizer.end_token_id,
            pad_value=self.tokenizer.pad_token_id,
            sequence_length=self.sequence_length,
            return_padding_mask=True,
        )
        self.built = True

    @preprocessing_function
    def call(
        self,
        x,
        y=None,
        sample_weight=None,
        sequence_length=None,
    ):
        sequence_length = sequence_length or self.sequence_length
        token_ids, padding_mask = self.packer(
            self.tokenizer(x),
            sequence_length=sequence_length,
            add_start_value=self.add_start_token,
            add_end_value=self.add_end_token,
        )
        x = {
            "token_ids": token_ids,
            "padding_mask": padding_mask,
        }
        return keras.utils.pack_x_y_sample_weight(x, y, sample_weight)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "sequence_length": self.sequence_length,
                "add_start_token": self.add_start_token,
                "add_end_token": self.add_end_token,
            }
        )
        return config



import contextlib

from keras_hub.src.utils.preset_utils import SAFETENSOR_CONFIG_FILE
from keras_hub.src.utils.preset_utils import SAFETENSOR_FILE
from keras_hub.src.utils.preset_utils import check_file_exists
from keras_hub.src.utils.preset_utils import get_file
from keras_hub.src.utils.preset_utils import load_json

try:
    import safetensors
except ImportError:
    safetensors = None


class SafetensorLoader(contextlib.ExitStack):
    def __init__(self, preset, prefix=None, fname=None):
        super().__init__()

        if safetensors is None:
            raise ImportError(
                "Converting from the huggingface/transformers model format"
                "requires the safetensors package."
                "Please install with `pip install safetensors`."
            )

        self.preset = preset
        if check_file_exists(preset, SAFETENSOR_CONFIG_FILE):
            self.safetensor_config = load_json(preset, SAFETENSOR_CONFIG_FILE)
        else:
            self.safetensor_config = None
        self.safetensor_files = {}
        self.prefix = prefix

        if fname is not None and self.safetensor_config is not None:
            raise ValueError(
                f"Cannot specify `fname` if {SAFETENSOR_CONFIG_FILE} exists. "
                f"Received: fname={fname}"
            )
        self.fname = fname  # Specify the name of the safetensor file.

    def get_prefixed_key(self, hf_weight_key, dict_like):
        """
        Determine and return a prefixed key for a given hf weight key.

        This method checks if there's a common prefix for the weight keys and caches it
        for future use.

        Args:
            hf_weight_key (str): The hf weight key to check for a prefix.
            dict_like (object): An object to get keys of safetensor file using keys() method.

        Returns:
            str: The full key including the prefix (if any).
        """
        if self.prefix is not None:
            return self.prefix + hf_weight_key

        for full_key in dict_like.keys():
            if full_key.endswith(hf_weight_key) and full_key != hf_weight_key:
                self.prefix = full_key[: -len(hf_weight_key)]
                return full_key

        self.prefix = ""
        return hf_weight_key

    def get_tensor(self, hf_weight_key):
        if self.safetensor_config is None:
            fname = self.fname if self.fname is not None else SAFETENSOR_FILE
        else:
            full_key = self.get_prefixed_key(
                hf_weight_key, self.safetensor_config["weight_map"]
            )
            fname = self.safetensor_config["weight_map"][full_key]

        if fname in self.safetensor_files:
            file = self.safetensor_files[fname]
        else:
            path = get_file(self.preset, fname)
            file = self.enter_context(
                safetensors.safe_open(path, framework="np")
            )
            self.safetensor_files[fname] = file

        full_key = self.get_prefixed_key(hf_weight_key, file)
        return file.get_tensor(full_key)

    def port_weight(self, keras_variable, hf_weight_key, hook_fn=None):
        hf_tensor = self.get_tensor(hf_weight_key)
        if hook_fn:
            hf_tensor = hook_fn(hf_tensor, list(keras_variable.shape))
        keras_variable.assign(hf_tensor)



from keras_hub.src.models.t5.t5_backbone import T5Backbone
from keras_hub.src.models.t5.t5_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, T5Backbone)



"""Llama 3 model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "llama3_8b_en": {
        "metadata": {
            "description": "8 billion parameter, 32-layer, base LLaMA 3 model.",
            "params": 8030261248,
            "official_name": "LLaMA 3",
            "path": "llama3",
            "model_card": "https://github.com/meta-llama/llama3",
        },
        "kaggle_handle": "kaggle://keras/llama3/keras/llama3_8b_en/3",
    },
    "llama3_8b_en_int8": {
        "metadata": {
            "description": (
                "8 billion parameter, 32-layer, base LLaMA 3 model with "
                "activation and weights quantized to int8."
            ),
            "params": 8031894016,
            "official_name": "LLaMA 3",
            "path": "llama3",
            "model_card": "https://github.com/meta-llama/llama3",
        },
        "kaggle_handle": "kaggle://keras/llama3/keras/llama3_8b_en_int8/1",
    },
    "llama3_instruct_8b_en": {
        "metadata": {
            "description": (
                "8 billion parameter, 32-layer, instruction tuned LLaMA 3 "
                "model."
            ),
            "params": 8030261248,
            "official_name": "LLaMA 3",
            "path": "llama3",
            "model_card": "https://github.com/meta-llama/llama3",
        },
        "kaggle_handle": "kaggle://keras/llama3/keras/llama3_instruct_8b_en/3",
    },
    "llama3_instruct_8b_en_int8": {
        "metadata": {
            "description": (
                "8 billion parameter, 32-layer, instruction tuned LLaMA 3 "
                "model with activation and weights quantized to int8."
            ),
            "params": 8031894016,
            "official_name": "LLaMA 3",
            "path": "llama3",
            "model_card": "https://github.com/meta-llama/llama3",
        },
        "kaggle_handle": (
            "kaggle://keras/llama3/keras/llama3_instruct_8b_en_int8/1"
        ),
    },
}



import keras
from keras import ops

from keras_hub.src.models.sam.sam_layers import (
    MultiHeadAttentionWithDownsampling,
)
from keras_hub.src.models.sam.sam_layers import TwoWayMultiHeadAttention


class TwoWayTransformer(keras.layers.Layer):
    """A two-way cross-attention transformer decoder.

    A transformer decoder that attends to an input image using
    queries whose positional embedding is supplied.
    The transformer decoder design is shown in
    [1](https://arxiv.org/abs/2304.02643).
    Each decoder layer performs 4 steps:
    (1) self-attention on the tokens,
    (2) cross-attention from tokens (as queries) to the image embedding,
    (3) a point-wise MLPupdates each token, and
    (4) cross-attention from the image embedding (as
    queries) to tokens. This last step updates the image embedding with prompt
    information. Each self/cross-attention and MLP has a residual connection
    and layer normalization.
    To ensure the decoder has access to critical geometric information the
    positional encodings are added to the image embedding whenever they
    participate in an attention layer. Additionally, the entire original
    prompt tokens (including their positional encodings) are re-added to the
    updated tokens whenever they participate in an attention layer. This
    allows for a strong dependence on both the prompt token's geometric
    location and type.

    Args:
        num_layers: int, optional. The num_layers of the attention blocks (the number
            of attention blocks to use). Defaults to `2`.
        hidden_size: int, optional. The number of features of the input image
            and point embeddings. Defaults to `256`.
        num_heads: int, optional. Number of heads to use in the attention
            layers. Defaults to `8`.
        intermediate_dim: int, optional. The number of units in the hidden layer of
            the MLP block used in the attention layers. Defaults to `2048`.
        activation: str, optional. The activation of the MLP block's output
            layer used in the attention layers. Defaults to `"relu"`.
        attention_downsample_rate: int, optional. The downsample rate of the
            attention layers. Defaults to `2`.
    """

    def __init__(
        self,
        *,
        num_layers=2,
        hidden_size=256,
        num_heads=8,
        intermediate_dim=2048,
        activation="relu",
        attention_downsample_rate=2,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        self.intermediate_dim = intermediate_dim
        self.activation = activation
        self.attention_downsample_rate = attention_downsample_rate
        self.layers = []
        for i in range(num_layers):
            self.layers.append(
                TwoWayMultiHeadAttention(
                    num_heads=num_heads,
                    key_dim=hidden_size // num_heads,
                    intermediate_dim=intermediate_dim,
                    skip_first_layer_pos_embedding=(i == 0),
                    attention_downsample_rate=attention_downsample_rate,
                    activation=activation,
                    dtype=self.dtype_policy,
                )
            )
        self.final_attention_token_to_image = (
            MultiHeadAttentionWithDownsampling(
                num_heads=num_heads,
                key_dim=hidden_size // num_heads,
                downsample_rate=attention_downsample_rate,
                dtype=self.dtype_policy,
            )
        )
        self.final_layer_norm = keras.layers.LayerNormalization(
            epsilon=1e-5, dtype=self.dtype_policy
        )

    def build(self, input_shape=None):
        for layer in self.layers:
            layer.build()
        self.final_attention_token_to_image.build()
        self.final_layer_norm.build([None, None, self.hidden_size])
        self.built = True

    def call(
        self, image_embedding, image_positional_embeddings, point_embedding
    ):
        shape = ops.shape(image_embedding)
        B, H, W, C = shape[0], shape[1], shape[2], shape[3]
        image_embedding = ops.reshape(image_embedding, (B, H * W, C))

        shape = ops.shape(image_positional_embeddings)
        B, H, W, C = shape[0], shape[1], shape[2], shape[3]
        image_positional_embeddings = ops.reshape(
            image_positional_embeddings, (B, H * W, C)
        )
        queries = point_embedding
        keys = image_embedding

        for layer in self.layers:
            queries, keys = layer(
                queries=queries,
                keys=keys,
                query_pos_embedding=point_embedding,
                key_pos_embedding=image_positional_embeddings,
            )

        queries_with_pos_embedding = queries + point_embedding
        keys_with_pos_embedding = keys + image_positional_embeddings
        attention_map = self.final_attention_token_to_image(
            query=queries_with_pos_embedding,
            key=keys_with_pos_embedding,
            value=keys,
        )
        queries = queries + attention_map
        queries = self.final_layer_norm(queries)

        return queries, keys

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_layers": self.num_layers,
                "hidden_size": self.hidden_size,
                "num_heads": self.num_heads,
                "intermediate_dim": self.intermediate_dim,
                "activation": self.activation,
                "attention_downsample_rate": self.attention_downsample_rate,
            }
        )
        return config



import keras
from keras import ops

from keras_hub.src.bounding_box.converters import _encode_box_to_deltas
from keras_hub.src.bounding_box.iou import compute_iou
from keras_hub.src.models.retinanet.anchor_generator import AnchorGenerator
from keras_hub.src.models.retinanet.box_matcher import BoxMatcher
from keras_hub.src.utils import tensor_utils


class RetinaNetLabelEncoder(keras.layers.Layer):
    """Transforms the raw labels into targets for training.

    RetinaNet is a single-stage object detection network that uses a feature
    pyramid network and focal loss. This class is crucial for preparing the
    ground truth data to match the network's anchor-based detection approach.

    This class generates targets for a batch of samples which consists of input
    images, bounding boxes for the objects present, and their class ids. It
    matches ground truth boxes to anchor boxes based on IoU (Intersection over
    Union) and encodes the box coordinates as offsets from the anchors.

    Targets are always represented in 'center_yxwh' format for numerical
    consistency during training, regardless of the input format.

    Args:
        bounding_box_format: str. The format of bounding boxes of input dataset.
            Refer TODO: Add link to Keras Core Docs.
        min_level: int. Minimum level of the output feature pyramid.
        max_level: int. Maximum level of the output feature pyramid.
        num_scales: int. Number of intermediate scales added on each level.
            For example, num_scales=2 adds one additional intermediate anchor
            scale [2^0, 2^0.5] on each level.
        aspect_ratios: List[float]. Aspect ratios of anchors added on
            each level. Each number indicates the ratio of width to height.
        anchor_size: float. Scale of size of the base anchor relative to the
            feature stride 2^level.
        positive_threshold:  float. the threshold to set an anchor to positive
            match to gt box. Values above it are positive matches.
            Defaults to `0.5`
        negative_threshold: float. the threshold to set an anchor to negative
            match to gt box. Values below it are negative matches.
            Defaults to `0.4`
        box_variance: List[float]. The scaling factors used to scale the
            bounding box targets.
            Defaults to `[0.1, 0.1, 0.2, 0.2]`.
        background_class: int. The class ID used for the background class,
            Defaults to `-1`.
        ignore_class: int. The class ID used for the ignore class,
            Defaults to `-2`.
        box_matcher_match_values: List[int]. Representing
            matched results (e.g. positive or negative or ignored match).
            `len(match_values)` must equal to `len(thresholds) + 1`.
            Defaults to `[-1, -2, -1]`.
        box_matcher_force_match_for_each_col: bool. If True, each column
            (ground truth box) will be matched to at least one row (anchor box).
            This means some columns may be matched to multiple rows while others
            may not be matched to any.
            Defaults to `False`.

    Note: `tf.RaggedTensor` are not supported.
    """

    def __init__(
        self,
        bounding_box_format,
        min_level,
        max_level,
        num_scales,
        aspect_ratios,
        anchor_size,
        positive_threshold=0.5,
        negative_threshold=0.4,
        box_variance=[0.1, 0.1, 0.2, 0.2],
        background_class=-1.0,
        ignore_class=-2.0,
        box_matcher_match_values=[-1, -2, 1],
        box_matcher_force_match_for_each_col=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.bounding_box_format = bounding_box_format
        self.min_level = min_level
        self.max_level = max_level
        self.num_scales = num_scales
        self.aspect_ratios = aspect_ratios
        self.anchor_size = anchor_size
        self.positive_threshold = positive_threshold
        self.box_variance = box_variance
        self.negative_threshold = negative_threshold
        self.background_class = background_class
        self.ignore_class = ignore_class

        self.anchor_generator = AnchorGenerator(
            bounding_box_format=bounding_box_format,
            min_level=min_level,
            max_level=max_level,
            num_scales=num_scales,
            aspect_ratios=aspect_ratios,
            anchor_size=anchor_size,
        )

        self.box_matcher = BoxMatcher(
            thresholds=[negative_threshold, positive_threshold],
            match_values=box_matcher_match_values,
            force_match_for_each_col=box_matcher_force_match_for_each_col,
        )

    def build(self, images_shape, gt_boxes_shape, gt_classes_shape):
        self.built = True

    def call(self, images, gt_boxes, gt_classes):
        """Creates box and classification targets for a batch.

        Args:
            images: A Tensor. The  input images argument should be
                of shape `[B, H, W, C]` or `[B, C, H, W]`.
            gt_boxes: A Tensor with shape of `[B, num_boxes, 4]`.
            gt_labels: A Tensor with shape of `[B, num_boxes, num_classes]`

        Returns:
            box_targets: A Tensor of shape `[batch_size, num_anchors, 4]`
                containing the encoded box targets.
            class_targets: A Tensor of shape `[batch_size, num_anchors, 1]`
                containing the class targets for each anchor.
        """

        images_shape = ops.shape(images)
        if len(images_shape) != 4:
            raise ValueError(
                "`RetinaNetLabelEncoder`'s `call()` method does not "
                "support  unbatched inputs for the `images` argument. "
                f"Received `shape(images)={images_shape}`."
            )
        image_shape = images_shape[1:]

        if len(ops.shape(gt_classes)) == 2:
            gt_classes = ops.expand_dims(gt_classes, axis=-1)

        anchor_boxes = self.anchor_generator(images)
        anchor_boxes = ops.concatenate(list(anchor_boxes.values()), axis=0)

        box_targets, class_targets = self._encode_sample(
            gt_boxes, gt_classes, anchor_boxes, image_shape
        )
        box_targets = ops.reshape(
            box_targets, (-1, ops.shape(box_targets)[1], 4)
        )
        return box_targets, class_targets

    def _encode_sample(self, gt_boxes, gt_classes, anchor_boxes, image_shape):
        """Creates box and classification targets for a batched sample.

        Matches ground truth boxes to anchor boxes based on IOU.
        1. Calculates the pairwise IOU for the M `anchor_boxes` and N `gt_boxes`
            to get a `(M, N)` shaped matrix.
        2. The ground truth box with the maximum IOU in each row is assigned to
            the anchor box provided the IOU is greater than `match_iou`.
        3. If the maximum IOU in a row is less than `ignore_iou`, the anchor
            box is assigned with the background class.
        4. The remaining anchor boxes that do not have any class assigned are
            ignored during training.

        Args:
            gt_boxes: A Tensor of shape `[B, num_boxes, 4]`. Should be in
                `bounding_box_format`.
            gt_classes: A Tensor fo shape `[B, num_boxes, num_classes, 1]`.
            anchor_boxes: A Tensor with the shape `[total_anchors, 4]`
                representing all the anchor boxes for a given input image shape,
                where each anchor box is of the format `[x, y, width, height]`.
            image_shape: Tuple indicating the image shape `[H, W, C]`.

        Returns:
            Encoded boudning boxes in the format of `center_yxwh` and
            corresponding labels for each encoded bounding box.
        """

        iou_matrix = compute_iou(
            anchor_boxes,
            gt_boxes,
            bounding_box_format=self.bounding_box_format,
            image_shape=image_shape,
        )

        matched_gt_idx, matched_vals = self.box_matcher(iou_matrix)
        matched_vals = ops.expand_dims(matched_vals, axis=-1)
        positive_mask = ops.cast(ops.equal(matched_vals, 1), self.dtype)
        ignore_mask = ops.cast(ops.equal(matched_vals, -2), self.dtype)

        matched_gt_boxes = tensor_utils.target_gather(gt_boxes, matched_gt_idx)

        matched_gt_boxes = ops.reshape(
            matched_gt_boxes, (-1, ops.shape(matched_gt_boxes)[1], 4)
        )

        box_target = _encode_box_to_deltas(
            anchors=anchor_boxes,
            boxes=matched_gt_boxes,
            anchor_format=self.bounding_box_format,
            box_format=self.bounding_box_format,
            variance=self.box_variance,
            image_shape=image_shape,
        )

        matched_gt_cls_ids = tensor_utils.target_gather(
            gt_classes, matched_gt_idx
        )
        cls_target = ops.where(
            ops.not_equal(positive_mask, 1.0),
            self.background_class,
            matched_gt_cls_ids,
        )
        cls_target = ops.where(
            ops.equal(ignore_mask, 1.0), self.ignore_class, cls_target
        )
        label = ops.concatenate(
            [box_target, ops.cast(cls_target, box_target.dtype)], axis=-1
        )

        # In the case that a box in the corner of an image matches with an all
        # -1 box that is outside the image, we should assign the box to the
        # ignore class. There are rare cases where a -1 box can be matched,
        # resulting in a NaN during training. The unit test passing all -1s to
        # the label encoder ensures that we properly handle this edge-case.
        label = ops.where(
            ops.expand_dims(ops.any(ops.isnan(label), axis=-1), axis=-1),
            self.ignore_class,
            label,
        )

        return label[:, :, :4], label[:, :, 4]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "bounding_box_format": self.bounding_box_format,
                "min_level": self.min_level,
                "max_level": self.max_level,
                "num_scales": self.num_scales,
                "aspect_ratios": self.aspect_ratios,
                "anchor_size": self.anchor_size,
                "positive_threshold": self.positive_threshold,
                "box_variance": self.box_variance,
                "negative_threshold": self.negative_threshold,
                "background_class": self.background_class,
                "ignore_class": self.ignore_class,
            }
        )
        return config

    def compute_output_shape(
        self, images_shape, gt_boxes_shape, gt_classes_shape
    ):
        min_level = self.anchor_generator.min_level
        max_level = self.anchor_generator.max_level
        batch_size, image_H, image_W = images_shape[:-1]

        total_num_anchors = 0
        for i in range(min_level, max_level + 1):
            total_num_anchors += (
                (image_H // 2 ** (i))
                * (image_W // 2 ** (i))
                * self.anchor_generator.anchors_per_location
            )

        return (batch_size, total_num_anchors, 4), (
            batch_size,
            total_num_anchors,
        )
